/*! \file ddag.hpp
 *  \brief Definitions of the object DDAG used by the user. Three template specializations are available.
 *  DDAG is the structure of the tree, not data. data are stored in DProperty objects.
 *  DDAG is an interface to the DDAG_impl
 */
#ifndef DDAG_H
#define DDAG_H

#include "ddag_impl.hpp"

namespace skelgis{
  
  //-------------------------------------------------------------------------------
  //!  base template class of Distributed DAG
  /*!
    This template class defines the base distributed dag
    This basic template does not have an implementation, only specializations have
    direction of the DAG is not usefull the input file of the user precise the direction
    no parameter on the physical border because roots and leafs nodes are always managed in a different way than standard elements (different neighborhood etc.)
    \tparam node is a boolean parameter to indicate if computations on nodes are needed for this DDAG
    \tparam edge is a boolean parameter to indicate if computations on edges are needed for this DDAG
  */
  //-------------------------------------------------------------------------------
  template<bool node, bool edge> struct DDAG;
  //-------------------------------------------------------------------------------
  //!  First specialization of the DDAG class
  /*!
    This is the more complex case where both nodes and edges have to be computed
  */
  //-------------------------------------------------------------------------------
  template<> struct DDAG<true,true>
  //-------------------------------------------------------------------------------
  {
  protected:
    DDAG_impl<true,true> * ddag;
  public:
    DDAG(const char * file)
    {
      ddag = new DDAG_impl<true,true>(file);
    }
    ~DDAG()
    {
      delete ddag;
    }
    
    static bool const nod = true;
    static bool const edg = true;

    inline unsigned int getIndexEdge(unsigned int ind){return ddag->getIndexEdge(ind);}
    //-------------------------------------------------------------------------------
    //! get the number of nodes
    /*!
      \return dim_node the number of nodes in the local data structure
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimNode(){return ddag->getGDimNode();}
    //-------------------------------------------------------------------------------
    //! get the number of edges
    /*!
      \return dim_edge the number of edges in the local data structure
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimEdge(){return ddag->getGDimEdge();}
    //-------------------------------------------------------------------------------

    //-------------------------------------------------------------------------------
    //! get the pointer to the DDAG_impl
    /*!
      \return the pointer to the DDAG_impl
    */
    //-------------------------------------------------------------------------------
    DDAG_impl<true,true> * getDDAG_impl(){return ddag;}
    //-------------------------------------------------------------------------------
    unsigned int getSizeEdges(){return ddag->getDimEdge();}
    unsigned int getSizeNodes(){return ddag->getDimNode();}
  };
  //-------------------------------------------------------------------------------
  //!  Second specialization of the DDAG class
  /*!
    Computations on edges only
  */
  //-------------------------------------------------------------------------------
  template<> struct DDAG<false,true>
  //-------------------------------------------------------------------------------
  {
  protected:
    DDAG_impl<false,true> * ddag;
  public:
    DDAG(const char * file)
    {
      ddag = new DDAG_impl<false,true>(file);
    }
    ~DDAG()
    {
      delete ddag;
    }
    
    static bool const nod = false;
    static bool const edg = true;

    DDAG_impl<false,true> * getDDAG_impl(){return ddag;}
  };
  //-------------------------------------------------------------------------------
  //!  Third specialization of the DDAG class
  /*!
    Computations on nodes only
  */
  //-------------------------------------------------------------------------------
  template<> struct DDAG<true,false>
  //-------------------------------------------------------------------------------
  {
  protected:
    DDAG_impl<true,false> * ddag;
  public:
    DDAG(const char * file)
    {
      ddag = new DDAG_impl<true,false>(file);
    }
    ~DDAG()
    {
      delete ddag;
    }

    static bool const nod = true;
    static bool const edg = false;

    DDAG_impl<true,false> * getDDAG_impl(){return ddag;}
  };
  //-------------------------------------------------------------------------------
};

#endif
