/*! \file pmap_nodes_impl.hpp
 *  \brief Definitions of the object PMap_Nodes_impl implementation to map data to the nodes of a DDAG. Three template specializations are available.
 */
#ifndef DPMAP_NODES_IMPL_H
#define DPMAP_NODES_IMPL_H

#include "ddag.hpp"
#include "../../iterators/DDAG/iterator.hpp"
#include "../../util/communications.hpp"
#include "dpmap_edges.hpp"
#include<vector>

namespace skelgis{

  //-------------------------------------------------------------------------------
  //!  base template class of Property map on nodes
  /*!
    This template class defines the base PMap_Nodes_impl
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam overlap is the overlap size needed to compute the data

    for a standard type T (two first specializations) the overlap can be 0 or 1 and the size is the default one = 1
    for a pointer type T (two second specializations) the overlap can be any value and the size is > 1 because a table is gonna be mapped to each node
  */
  //-------------------------------------------------------------------------------
  template<class DD,class T, int overlap> struct DPMap_Nodes_impl;
  //-------------------------------------------------------------------------------
  //!  first specialization of DPMap_Nodes_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The overlap is specialized to the value 1
  */
  //-------------------------------------------------------------------------------
  template<class DD,class T> struct DPMap_Nodes_impl<DD,T,1>
  //-------------------------------------------------------------------------------
  {
  public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;
    /*!< object to make MPI communications */
    Communications<T> * comm;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on nodes (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      //parallel read to write CF DMatrix
      FILE *pmapfile=fopen(file,"r");

      T any=0;
      for(unsigned int i=0;i<dim_node+dim_tor;i++)
	{
	  for(unsigned int j=0;j<ddag->getIndexNode(i);j++)
	    int read = fscanf(pmapfile,"%lf", &any);
	  int read = fscanf(pmapfile,"%lf", &data[i]);
	  int seek = fseek(pmapfile,0,SEEK_SET);
	}

      fclose(pmapfile);
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on nodes (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      dim_node = ddag->getDimNode();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      data = new T[dim_node+dim_tor]();
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate nodes to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      dim_node = ddag->getDimNode();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      data = new T[dim_node+dim_tor]();
      for(unsigned int i=0;i<dim_node+dim_tor;i++)
	{
	  data[i] = val;
	}
    }
    //-------------------------------------------------------------------------------
    //!  init function for MPI communications
    /*!
    */
    //-------------------------------------------------------------------------------
    void init_comm()
    {
      //init the communication object
      comm = new Communications<T>();
    }
    //-------------------------------------------------------------------------------
    
  public:
    /*!< dim_node is the number of local nodes */
    unsigned int dim_node;
    /*!< dim_tor is the number of additionnal nodes to receive */
    unsigned int dim_tor;
    /*!< data mapped on nodes of the ddag */
    T * data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on nodes (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes_impl(DD &dag, const char * file)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //communication initialization
      init_comm();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Nodes
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of nodes for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes_impl(DD &dag, T val)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //communication initialization
      init_comm();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Nodes
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Nodes_impl()
    //-------------------------------------------------------------------------------
    {
      delete [] data;
      delete comm;
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on roots nodes
    /*!
      \return the begin iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_roots(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs());}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on roots nodes
    /*!
      \return the end iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_roots(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs()+ddag->getDimRoots());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on leafs nodes
    /*!
      \return the begin iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_leafs(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm());}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on leafs nodes
    /*!
      \return the end iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_leafs(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes with communication with other processors
    /*!
      \return the begin iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_comm(){return iterator_dag(ddag->getDimLoc());}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes with communication with other processors
    /*!
      \return the end iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_comm(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes without communication with other processors
    /*!
      \return the begin iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_loc(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes without communication with other processors
    /*!
      \return the end iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_loc(){return iterator_dag(ddag->getDimLoc());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the begin iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_in(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the end iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_in(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on all nodes (roots, leafs, comm and loc)
    /*!
      \return the begin iterator on all nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_all(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on all nodes (roots, leafs, comm and loc)
    /*!
      \return the end iterator on all nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_all(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs()+ddag->getDimRoots());}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input edges for the node it
    /*!
      \param it is the iterator on the node to get input edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInEdges(iterator_dag it){return ddag->getInEdges(it._rank);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output edges for the node it
    /*!
      \param it is the iterator on the node to get output edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutEdges(iterator_dag it){return ddag->getOutEdges(it._rank);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input nodes for the node it
    /*!
      \param it is the iterator on the node to get input nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInNodes(iterator_dag it){return ddag->getInNodes(it._rank);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output nodes for the node it
    /*!
      \param it is the iterator on the node to get output nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutNodes(iterator_dag it){return ddag->getOutNodes(it._rank);}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_nodes property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //prepare data to send
	  unsigned int nb_toSend = ddag->getNbNodesToSend(j);
	  if(nb_toSend>0)
	    {
	      T * toSend = (T*)calloc(nb_toSend,sizeof(T));
	      unsigned int * indexes = ddag->getNodesToSend(j);
	      for(unsigned int i=0;i<nb_toSend;i++)
		toSend[i] = data[indexes[i]];

	      comm->ISend(toSend, nb_toSend, j);

	      free(indexes);
	      free(toSend);
	    }
	  //prepare data to receive
	  unsigned int nb_toRecv = ddag->getNbNodesToRecv(j);
	  if(nb_toRecv>0)
	    comm->IRecv(nb_toRecv, j);
	 
	}
    }
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_nodes property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_nodes property
    */
    //-------------------------------------------------------------------------------
    void end_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //prepare data to send
	  unsigned int nb_toSend = ddag->getNbNodesToSend(j);
	  if(nb_toSend>0)
	    comm->Wait_s(j);
	  unsigned int nb_toRecv = ddag->getNbNodesToRecv(j);
	  if(nb_toRecv>0)
	    {
	      T * toRecv = (T*)calloc(nb_toRecv,sizeof(T));
	      comm->Wait_r(toRecv,j,nb_toRecv);
	      unsigned int * indexes = ddag->getNodesToRecv(j);
	      //change values in data
	      for(int k=0;k<nb_toRecv;k++)
		data[indexes[k]] = toRecv[k];
	      free(indexes);
	      free(toRecv);
	    }
	}
    }
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------
  //!  second specialization of DPMap_Nodes_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The overlap is specialized to the value 0
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Nodes_impl<DD,T,0>
  {
    public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on nodes (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      //parallel read to write CF DMatrix
      FILE *pmapfile=fopen(file,"r");

      T any=0;
      for(unsigned int i=0;i<dim_node;i++)
	{
	  for(unsigned int j=0;j<ddag->getIndexNode(i);j++)
	    int read = fscanf(pmapfile,"%lf", &any);
	  int read = fscanf(pmapfile,"%lf", &data[i]);
	  int seek = fseek(pmapfile,0,SEEK_SET);
	}

      fclose(pmapfile);
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on nodes (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      dim_node = ddag->getDimNode();
      data = new T[dim_node]();
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate nodes to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      dim_node = ddag->getDimNode();
      data = new T[dim_node]();
      for(unsigned int i=0;i<dim_node;i++)
	{
	  data[i] = val;
	}
    }
    
  public:
    /*!< dim_node is the number of local nodes */
    unsigned int dim_node;
    /*!< data mapped on nodes of the ddag */
    T * data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on nodes (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes_impl(DD &dag, const char * file)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Nodes
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of nodes for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes_impl(DD &dag, T val)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Nodes
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Nodes_impl()
    //-------------------------------------------------------------------------------
    {
      delete [] data;
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on roots nodes
    /*!
      \return the begin iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_roots(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs());}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on roots nodes
    /*!
      \return the end iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_roots(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs()+ddag->getDimRoots());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on leafs nodes
    /*!
      \return the begin iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_leafs(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm());}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on leafs nodes
    /*!
      \return the end iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_leafs(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes with communication with other processors
    /*!
      \return the begin iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_comm(){return iterator_dag(ddag->getDimLoc());}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes with communication with other processors
    /*!
      \return the end iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_comm(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes without communication with other processors
    /*!
      \return the begin iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_loc(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes without communication with other processors
    /*!
      \return the end iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_loc(){return iterator_dag(ddag->getDimLoc());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the begin iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_in(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the end iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_in(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm());}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on all nodes (roots, leafs, comm and loc)
    /*!
      \return the begin iterator on all nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_all(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on all nodes (roots, leafs, comm and loc)
    /*!
      \return the end iterator on all nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_all(){return iterator_dag(ddag->getDimLoc()+ddag->getDimComm()+ddag->getDimLeafs()+ddag->getDimRoots());}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input edges for the node it
    //!  to get a vector of iterator on input edges for the node it
    /*!
      \param it is the iterator on the node to get input edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInEdges(iterator_dag it){return ddag->getInEdges(it._rank);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output edges for the node it
    /*!
      \param it is the iterator on the node to get output edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutEdges(iterator_dag it){return ddag->getOutEdges(it._rank);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input nodes for the node it
    /*!
      \param it is the iterator on the node to get input nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInNodes(iterator_dag it){return ddag->getInNodes(it._rank);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output nodes for the node it
    /*!
      \param it is the iterator on the node to get output nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutNodes(iterator_dag it){return ddag->getOutNodes(it._rank);}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_nodes property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    inline void start_communications(){}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_nodes property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_nodes property
    */
    //-------------------------------------------------------------------------------
    inline void end_communications(){}
    //-------------------------------------------------------------------------------

  };
  //-------------------------------------------------------------------------------
  //!  third specialization of DPMap_Nodes_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tapram overlap is the overlap size needed to compute the data

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The overlap is not specialized and is [0,...,size]
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T,int overlap> struct DPMap_Nodes_impl<DD,T*,overlap>
  {
  public :

  public:
    DPMap_Nodes_impl(){};
    ~DPMap_Nodes_impl(){};
  };
  //-------------------------------------------------------------------------------
  //!  fourth specialization of DPMap_Nodes_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tapram overlap is the overlap size needed to compute the data

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The overlap is specialized to the value 0
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Nodes_impl<DD,T*,0>
  {
  public :

  public:
    DPMap_Nodes_impl(){};
    ~DPMap_Nodes_impl(){};
  };

};


#endif
