/*! \file dmatrix_impl.hpp
 *  \brief Definitions of the distributed matrix implementation object. This class is not used by the user. Four template specializations are available as for DMatrix.
 */
#ifndef DMATRIX_IMPL_H
#define DMATRIX_IMPL_H

#include "dmatrix.hpp"
#include "../../util/utility.hpp"
#include "../../util/mpi_.hpp"
#include "../../util/communications.hpp"

#include "../../iterators/DMatrix/iterator.hpp"
#include "../../iterators/DMatrix/iterator_cont.hpp"
#include "../../iterators/DMatrix/iterator_rev.hpp"
#include "../../iterators/DMatrix/iterator_phb.hpp"
#include "../../iterators/DMatrix/iterator_step.hpp"
#include "../../iterators/DMatrix/iterator_line.hpp"

#include<math.h>

namespace skelgis{

  //================================================================================
  //!  base template Distributed Matrix Class
  /*!
    This template class defines the base distributed matrix. Every DMatrix_impl class inherit from this one.
    \tparam T is the type of data in the matrix
  */
  //-------------------------------------------------------------------------------
  template<class T> struct DMatrix_base
  //-------------------------------------------------------------------------------
  {
  public:
    //! data of the distributed matrix
    /*!
      In this table are stored the data of the distributed matrix.
    */
    T * data;
    //! Column and row positions for the current process in the general data distribution
    /*!
      The data distribution is made along height and/or width, each process is associated to a pair (col,row) in this data distribution.
      If the distribution is line=true, the col=0 for all processes.
    */
    int col,row;
    //! Total number of columns and rows in the general data distribution
    int cols,rows;
    //! Header defining the global matrix before parallel distribution
    HEADER head;
    //! Header defining the local matrix of the current process after data distribution
    HEADER loc_head;

    bool local; /*!< if the DMatrix is local to one process */

    protected:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix_base(){}
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix_base(){}
    //-------------------------------------------------------------------------------
    //! To get the height and the width of the last column and the last row
    /*!
      Because the data division along height and width are not always divisible by the number of processes, variables to remind the size of the last row and of the last column has to be defined.
    */
    int remainderh,remainderw;
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //!  DMatrix_impl class
  /*!
    template of the distributed matrix implementation of SkelGIS. This is the first specialization of the DMatrix_impl.
    \tparam T is the type of data to store in the DMatrix_impl
    \tparam R is the overlap distance needed by the calculation, in other words, the size of the physical border needed
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,bool line=true> struct DMatrix_impl : public DMatrix_base<T>
  //-------------------------------------------------------------------------------
  {
  public:

    HEADER border_head; /*!< Header defining the current local matrix with overlap R */

  protected:
    int mpi_ranks[2]; /*!< table of size 2 for the neighbor processes to make MPI exchanges */
    unsigned int beginPhBLeft,beginPhBRight,beginPhBUp,beginPhBDown; /*!< begining indexes for the physical borders iterators */
    unsigned int endPhBLeft,endPhBRight,endPhBUp,endPhBDown; /*!< ending indexes for the physical borders iterators */
    //-------------------------------------------------------------------------------
    //! Creation of the ditributed matrix
    /*!
      Data distribution and construction of the parallel data structure.
    */
    //-------------------------------------------------------------------------------
    void create()
    //-------------------------------------------------------------------------------
    {
      if(this->local==false)
	{
	  //division of the matrix in cols and rows
	  this->cols = 1;
	  this->rows = Mpi_::mpi_nb;

	  this->row=Mpi_::mpi_rank;
	  this->col=0;
      
	  this->loc_head.width=this->head.width;
      
	  this-> remainderh = 0;
	  //if the height of the matrix can be divided by the number of rows
	  if(this->head.height%this->rows==0)
	    this->loc_head.height=floor(this->head.height/this->rows);
	  //else we add one to the height of the submatrix and the last row will have less elements
	  else
	    {
	      this->remainderh = this->head.height - floor(this->head.height/this->rows)*this->rows;

	      if(this->row<this->remainderh)
		this->loc_head.height=floor(this->head.height/this->rows)+1;
	      else
		this->loc_head.height=floor(this->head.height/this->rows);
	    }

	  this->loc_head.x=this->head.x;
	  this->loc_head.y=this->head.y-this->row*this->loc_head.height*this->head.spacing;
	  this->loc_head.spacing=this->head.spacing;
	  this->loc_head.nodata=this->head.nodata;
	}
      else
	{
	  this->loc_head.x = this->head.x;
	  this->loc_head.y = this->head.y;
	  this->loc_head.width = this->head.width;
	  this->loc_head.height = this->head.height;
	  this->loc_head.spacing = this->head.spacing;
	  this->loc_head.nodata = this->head.nodata;

	  this->col=0;this->row=0;
	  this->cols=1;this->rows=1;
	  this-> remainderh = 0;
	}

      this->border_head.x = this->loc_head.x - R*this->loc_head.spacing;
      this->border_head.y = this->loc_head.y - R*this->loc_head.spacing;
      this->border_head.width = this->loc_head.width + R*2;
      this->border_head.height = this->loc_head.height + R*2;
      
      this->data = new T[this->border_head.width*this->border_head.height];
    }
    //-------------------------------------------------------------------------------
    //! Values initialization of the elements of the DMatrix_impl
    /*!
      \param value is the value to set to the elements
    */
    //-------------------------------------------------------------------------------
    void initValues(T value)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<this->border_head.width*this->border_head.height;i++)
	{
	  this->data[i]=value;
	}
    }
    //-------------------------------------------------------------------------------
    //! Calculation of the physical border limits
    //-------------------------------------------------------------------------------
    void phBLimits()
    //-------------------------------------------------------------------------------
    {
      beginPhBLeft = R*this->border_head.width;
      beginPhBRight = (R+1)*this->border_head.width-R;
      endPhBLeft = this->border_head.width*(this->border_head.height-R-1)+R;
      endPhBRight = (R+1+this->loc_head.height-1)*this->border_head.width;

      if(Mpi_::mpi_rank==0)
	{
	  beginPhBUp = R;
	  endPhBUp = R*this->border_head.width-R;
	}
      else
	{
	  beginPhBUp = 0;
	  endPhBUp = 0;
	}
      if(Mpi_::mpi_rank==Mpi_::mpi_nb-1)
	{
	  beginPhBDown = this->border_head.width*(this->loc_head.height+R)+R;
	  endPhBDown = this->border_head.width*this->border_head.height-R;
	}
      else
	{
	  beginPhBDown = 0;
	  endPhBDown = 0;
	}
    }
    //-------------------------------------------------------------------------------
    //! Read the zone of the binary file concerned for the current MPI process
    /*!
      \param binFile is binary file to read
    */
    //-------------------------------------------------------------------------------
    void read(const char * binFile)
    //-------------------------------------------------------------------------------
    {
      //read the zone of the file for my own block of data
      std::ifstream f(binFile, std::ios::binary | std::ios::in);

      //move to the begining of the bloc of data concerned
      unsigned int mvline;
      if(Mpi_::mpi_rank<this->remainderh)
	mvline = sizeof(HEADER) + (this->row*this->loc_head.height*this->head.width)*sizeof(T);
      else
	mvline = sizeof(HEADER) + ((this->remainderh*(this->loc_head.height+1)+(this->row-this->remainderh)*this->loc_head.height)*this->head.width)*sizeof(T);

      f.seekg(mvline,std::ios::beg);

      //read by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  for(unsigned int k=0;k<this->loc_head.width;k++)
	    {
	      T val;
	      unsigned int rank = R*this->border_head.width + j*this->border_head.width + k + R;
	      f.read(reinterpret_cast<char*>(&val),sizeof(T));
	      this->data[rank]=val;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! Get ranks
    /*!
      Get the ranks of the 2 neighbors for MPI exchanges
      \return 
      if the value is -55555 then there is no neighbor in this direction
      if the value is >0 it represent the MPI rank concerned by the exchange
    */
    //-------------------------------------------------------------------------------
    void getNeighborRanks()
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<2;i++)
	mpi_ranks[i] = -55555;
      //up
      if((this->row-1)>=0)
	mpi_ranks[0] = Mpi_::mpi_rank-1;
      //down
      if((this->row+1)<this->rows)
	mpi_ranks[1] = Mpi_::mpi_rank+1;
    }
    //-------------------------------------------------------------------------------
    //! Get the up border to send to neighbor processors
    /*!
      \return table of values to send
    */
    //-------------------------------------------------------------------------------
    T * getUpBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[this->loc_head.width*R];
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      unsigned int r1 = i*this->loc_head.width+j;
	      unsigned int r2 = (R*border_head.width + R) + i*border_head.width+j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the up border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setUpBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    this->data[(R) + i*border_head.width+j] = values[i*this->loc_head.width+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the down border to send to neighbor processors
    /*!
      \return table of values to send
    */
    //-------------------------------------------------------------------------------
    T * getDownBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[this->loc_head.width*R];
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      unsigned int r1 = i*this->loc_head.width+j;
	      unsigned int r2 = (border_head.height*border_head.width - 2*R*border_head.width + R) + i*border_head.width + j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the down border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setDownBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    this->data[(border_head.width*border_head.height - border_head.width*R + R) + i*border_head.width +j] = values[i*this->loc_head.width+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRight(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank+1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank-1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r--;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r - width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r + width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width -1 ;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r + width -1 ;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r - width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width - 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r - width - 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T right(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + 1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T left(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - 1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T up(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T down(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T rightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width +1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T leftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width -1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T rightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width +1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T leftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width -1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allX(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[b];
	  b++;
	}
      b++;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = this->data[b];
	  b++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allX(unsigned int rank, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
        {
          t[i] = this->data[b];
          b++;
        }
      b++;
      for(int i=R;i<2*R;i++)
        {
          t[i] = this->data[b];
          b++;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allY(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[b];
	  b = b + width;
	}
      b = b + width;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = this->data[b];
	  b = b + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allY(unsigned int rank,unsigned int width, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
        {
          t[i] = this->data[b];
          b = b + width;
        }
      b = b + width;
      for(int i=R;i<2*R;i++)
        {
          t[i] = this->data[b];
          b = b + width;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to leftUp, up, rightUp, right, rightDown, down, leftDown and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * eight(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = leftUp(rank,width);
      res[1] = up(rank,width);
      res[2] = rightUp(rank,width);
      res[3] = right(rank);
      res[4] = rightDown(rank,width);
      res[5] = down(rank,width);
      res[6] = leftDown(rank,width);
      res[7] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to up, right, down and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * four(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = up(rank,width);
      res[1] = right(rank);
      res[2] = down(rank,width);
      res[3] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T inright(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = rank%border_head.width;
      unsigned int r = rank + (R-re);
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T inleft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = border_head.width - rank%border_head.width -1 ;
      unsigned int r = rank - (R-re);
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T inup(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = border_head.height-floor(rank/border_head.width)-1;
      unsigned int r = rank - (R-re)*width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T indown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = floor(rank/border_head.width);
      unsigned int r = rank + (R-re)*width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the indexes associated to a precise rank
    /*!
      \param rank is the rank at which the indexes values are wanted
      \param col is the column return value
      \param li is the line return value
    */
    //-------------------------------------------------------------------------------
    void indexesIt(unsigned int rank,int &col,int &li)
    //-------------------------------------------------------------------------------
    {
      unsigned int relative_rank = rank - R*border_head.width - (floor(rank/border_head.width)-R)*R*2 - R;
      li = floor(relative_rank/this->loc_head.width);
      col = relative_rank%this->loc_head.width;
    }
    //-------------------------------------------------------------------------------

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix_impl() : DMatrix_base<T>(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      This constructor will simply allocate the good space.
      \param h is the header to use to build the matrix
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(HEADER h,const T value, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head=h;
      this->local=loc;
      create();
      initValues(value);
      getNeighborRanks();
      phBLimits();
    }
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      The constructor evaluate how to divide the matrix in submatrice and do the work.
      \param binFile is the path of the binary file you want to work on
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(const char * binFile, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head = Header::read(binFile);
      this->local=loc;
      create();
      initValues(0);
      read(binFile);
      getNeighborRanks();
      phBLimits();
    }
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix_impl()
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  delete [] this->data;
	  this->data = NULL;
	}
    }
    //-------------------------------------------------------------------------------
    //! Set value in the global middle
    /*!
      \param val is the value to affect
    */
    //-------------------------------------------------------------------------------
    void setGlobalMiddleValue(T val)
    //-------------------------------------------------------------------------------
    {
      if(this->row!=this->rows-1 || this->rows==1)
	{
	  unsigned int middle_h = floor(this->head.height/2)-1;
	  if(this->loc_head.height*this->row<=middle_h && this->loc_head.height*(this->row+1)>middle_h)
	    {
	      unsigned int j_loc = floor(this->head.width/2)-1;
	      unsigned int i_loc = middle_h-this->row*this->loc_head.height;
	      unsigned int rank = R*border_head.width +  i_loc*border_head.width + j_loc + R;
	      this->data[rank] = val;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! to set all the values of the physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setPhysicalBorder(T val)
    {
      iterator_phb_left<T,R> left =begin_phb_left();
      iterator_phb_left<T,R> leftEnd = end_phb_left();
      for(left;left<leftEnd;++left)
	this->data[left._rank] = val;
      iterator_phb_right<T,R> right =begin_phb_right();
      iterator_phb_right<T,R> rightEnd = end_phb_right();
      for(right;right<rightEnd;++right)
	this->data[right._rank] = val;
      iterator_phb_up<T,R> up =begin_phb_up();
      iterator_phb_up<T,R> upEnd = end_phb_up();
      for(up;up<upEnd;++up)
	this->data[up._rank] = val;
      iterator_phb_down<T,R> down =begin_phb_down();
      iterator_phb_down<T,R> downEnd = end_phb_down();
      for(down;down<downEnd;++down)
	this->data[down._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the right of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setRightPhysicalBorder(T val)
    {
      iterator_phb_right<T,R> it =begin_phb_right();
      iterator_phb_right<T,R> itEnd = end_phb_right();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the left of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setLeftPhysicalBorder(T val)
    {
      iterator_phb_left<T,R> it =begin_phb_left();
      iterator_phb_left<T,R> itEnd = end_phb_left();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the up of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setUpPhysicalBorder(T val)
    {
      iterator_phb_up<T,R> it =begin_phb_up();
      iterator_phb_up<T,R> itEnd = end_phb_up();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the down of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    void setDownPhysicalBorder(T val)
    {
      iterator_phb_down<T,R> it =begin_phb_down();
      iterator_phb_down<T,R> itEnd = end_phb_down();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> begin(){return iterator<T,R>(R*this->border_head.width+R,this->border_head.width);}
    inline iterator_cont<T,R> begin_cont(){return iterator_cont<T,R>(R*this->border_head.width+R,this->border_head.width);}
    inline iterator_rev<T,R> begin_rev(){return iterator_rev<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_step<T,R> begin_step(int step,int nb){return iterator_step<T,R>(R*this->border_head.width+R,this->border_head.width,step,nb);}
    inline iterator_phb_left<T,R> begin_phb_left(){return iterator_phb_left<T,R>(beginPhBLeft,this->border_head.width);}
    inline iterator_phb_right<T,R> begin_phb_right(){return iterator_phb_right<T,R>(beginPhBRight,this->border_head.width);}
    inline iterator_phb_up<T,R> begin_phb_up(){return iterator_phb_up<T,R>(beginPhBUp,this->border_head.width);}
    inline iterator_phb_down<T,R> begin_phb_down(){return iterator_phb_down<T,R>(beginPhBDown,this->border_head.width);}
    inline iterator_line<T,R> begin_line(){return iterator_line<T,R>(R*this->border_head.width+R,this->border_head.width);}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> end(){return iterator<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_cont<T,R> end_cont(){return iterator_cont<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_rev<T,R> end_rev(){return iterator_rev<T,R>(R*this->border_head.width+R,this->border_head.width);}
    inline iterator_step<T,R> end_step(){return iterator_step<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_phb_left<T,R> end_phb_left(){return iterator_phb_left<T,R>(endPhBLeft,this->border_head.width);}
    inline iterator_phb_right<T,R> end_phb_right(){return iterator_phb_right<T,R>(endPhBRight,this->border_head.width);}
    inline iterator_phb_up<T,R> end_phb_up(){return iterator_phb_up<T,R>(endPhBUp,this->border_head.width);}
    inline iterator_phb_down<T,R> end_phb_down(){return iterator_phb_down<T,R>(endPhBDown,this->border_head.width);}
    inline iterator_line<T,R> end_line(){return iterator_line<T,R>((this->border_head.width*this->border_head.height)-(R+1)*this->border_head.width+R,this->border_head.width);}
    //-------------------------------------------------------------------------------
    //! get the iterator on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator<T,R> getIterator(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = R*border_head.width + R + li*border_head.width+col;
	  return iterator<T,R>(rank,this->border_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator contiguous on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_cont<T,R> getIterator_cont(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = R*border_head.width + R + li*border_head.width+col;
	  return iterator_cont<T,R>(rank,this->border_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator reverse on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_rev<T,R> getIterator_rev(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = R*border_head.width + R + li*border_head.width+col;
	  return iterator_rev<T,R>(rank,this->border_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the indexes of the iterator on the matrix
    /*!
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    void getIndexes(iterator<T,R> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_cont<T,R> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_rev<T,R> it,int &col,int &li){indexesIt(it._rank,col,li);}
    //-------------------------------------------------------------------------------
    //! Print the matrix by bloc
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      std::stringstream st;
      for(unsigned int i=0;i<this->loc_head.height;i++)
        {
          for(unsigned int j=0;j<this->loc_head.width;j++)
            {
              st<<this->data[R*this->border_head.width + R + i*this->border_head.width+j]<<" ";
            }
          st<<"\n";
        }
      Mpi_::print(st.str());
    }
    //-------------------------------------------------------------------------------
    //! Print the matrix by bloc with borders
    //-------------------------------------------------------------------------------
    void printAll()
    //-------------------------------------------------------------------------------
    {
      std::stringstream st;
      for(unsigned int i=0;i<this->border_head.height;i++)
	{
	  for(unsigned int j=0;j<this->border_head.width;j++)
	    {
	      st<<this->data[i*this->border_head.width+j]<<" ";
	    }
	  st<<"\n";
	}
      Mpi_::print(st.str());
    }
    //-------------------------------------------------------------------------------
    //! Write the zone file concerned for the current MPI process
    /*!
      write the MPI prcess zone, and spread the data in the subMatrices OpenMP
      only one submatrice if OpenMP is disabled
      \param binFile is binary file to read
    */
    //-------------------------------------------------------------------------------
    void write(char * binFile)
    //-------------------------------------------------------------------------------
    {
      MPI_File fh;
      MPI_Status status;
      //open the output file
      MPI_File_open(MPI_COMM_WORLD, binFile,
		MPI_MODE_CREATE|MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);
   
      // process 0 writes the HEADER of the file
      if (Mpi_::mpi_rank==0)
	MPI_File_write_at(fh, 0,reinterpret_cast<char*>(&this->head) ,sizeof(HEADER) , MPI_CHAR, &status);
      //the begining of the bloc of write in
      unsigned int mvline;
      if (Mpi_::mpi_rank<this->remainderh)
	mvline=sizeof(HEADER)+this->row*this->loc_head.height*this->loc_head.width*sizeof(T);
      else
	mvline=sizeof(HEADER)+(this->row*this->loc_head.height+this->remainderh)*this->loc_head.width*sizeof(T);

      // pointer in the data of the dmatrix
      T *data_pointer=this->data+R*(this->border_head.width+1);

      //write by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  MPI_File_write_at(fh, mvline,reinterpret_cast<char*>(data_pointer) ,this->loc_head.width*sizeof(T) , MPI_CHAR, &status);
	  data_pointer+=this->border_head.width;
	  mvline+=this->loc_head.width*sizeof(T);
	}
      MPI_File_close(&fh);
    }
    //-------------------------------------------------------------------------------
    //! get borders
    /*!
      Get the borders from other processors
      MPI exchanges
    */
    //-------------------------------------------------------------------------------
    void getBorders()
    //-------------------------------------------------------------------------------
    {	  
      if(this->row!=0)
	{
	  T * toSend = getUpBorderToSend();
	  T * toGet = new T[this->loc_head.width*R];
	  Communications<T>::Exchanges(toSend,toGet,this->loc_head.width*R,mpi_ranks[0]);
	  setUpBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;

	}
      if(this->row!=(this->rows-1))
	{
	  T * toSend = getDownBorderToSend();
	  T * toGet = new T[this->loc_head.width*R];
	  Communications<T>::Exchanges(toSend,toGet,this->loc_head.width*R,mpi_ranks[1]);
	  setDownBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right of the current it.
      This method exists for each type of iterator with right values in SkelGIS.
      \param it is the iterator position
      \return a table of right values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRight(iterator<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_cont<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_step<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_rev<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_phb_left<T,R> it){return allRight(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left of the current it.
      This method exists for each type of iterator with left values in SkelGIS.
      \param it is the iterator position
      \return a table of left values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeft(iterator<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_cont<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_step<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_rev<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_phb_right<T,R> it){return allLeft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the up of the current it.
      This method exists for each type of iterator with up values in SkelGIS.
      \param it is the iterator position
      \return a table of up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllUp(iterator<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_cont<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_step<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_rev<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_phb_down<T,R> it){return allUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the down of the current it.
      This method exists for each type of iterator with down values in SkelGIS.
      \param it is the iterator position
      \return a table of down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllDown(iterator<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_cont<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_step<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_rev<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_phb_up<T,R> it){return allDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right down of the current it.
      This method exists for each type of iterator with right down values in SkelGIS.
      \param it is the iterator position
      \return a table of right down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightDown(iterator<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_cont<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_step<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_rev<T,R> it){return allRightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left down of the current it.
      This method exists for each type of iterator with left down values in SkelGIS.
      \param it is the iterator position
      \return a table of left down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftDown(iterator<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_cont<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_step<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_rev<T,R> it){return allLeftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right up of the current it.
      This method exists for each type of iterator with right up values in SkelGIS.
      \param it is the iterator position
      \return a table of right up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightUp(iterator<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_cont<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_step<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_rev<T,R> it){return allRightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left up of the current it.
      This method exists for each type of iterator with left up values in SkelGIS.
      \param it is the iterator position
      \return a table of left up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftUp(iterator<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_cont<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_step<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_rev<T,R> it){return allLeftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right value in SkelGIS.
      \param it is the iterator position
      \return the right neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRight(iterator<T,R> it){return right(it._rank);}
    inline T getRight(iterator_cont<T,R> it){return right(it._rank);}
    inline T getRight(iterator_step<T,R> it){return right(it._rank);}
    inline T getRight(iterator_rev<T,R> it){return right(it._rank);}
    inline T getRight(iterator_phb_left<T,R> it){return right(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left value in SkelGIS.
      \param it is the iterator position
      \return the left neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeft(iterator<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_cont<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_step<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_rev<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_phb_right<T,R> it){return left(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with up value in SkelGIS.
      \param it is the iterator position
      \return the up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getUp(iterator<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_cont<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_step<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_rev<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_phb_down<T,R> it){return up(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with down value in SkelGIS.
      \param it is the iterator position
      \return the down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getDown(iterator<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_cont<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_step<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_rev<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_phb_up<T,R> it){return down(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right down value in SkelGIS.
      \param it is the iterator position
      \return the right down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightDown(iterator<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_cont<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_step<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_rev<T,R> it){return rightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left down value in SkelGIS.
      \param it is the iterator position
      \return the left down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftDown(iterator<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_cont<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_step<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_rev<T,R> it){return leftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right up value in SkelGIS.
      \param it is the iterator position
      \return the right up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightUp(iterator<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_cont<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_step<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_rev<T,R> it){return rightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left up value in SkelGIS.
      \param it is the iterator position
      \return the left up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftUp(iterator<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_cont<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_step<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_rev<T,R> it){return leftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline T * getAllX(iterator<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_cont<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_step<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_rev<T,R> it){return allX(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline void getAllX(iterator<T,R> it, T * t){return allX(it,t);}
    inline void getAllX(iterator_cont<T,R> it, T * t){return allX(it,t);}
    inline void getAllX(iterator_rev<T,R> it, T * t){return allX(it,t);}
    inline void getAllX(iterator_step<T,R> it, T * t){return allX(it,t);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline T * getAllY(iterator<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_cont<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_step<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_rev<T,R> it){return allY(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline void getAllY(iterator<T,R> it, T * t){return allY(it,t);}
    inline void getAllY(iterator_cont<T,R> it, T * t){return allY(it,t);}
    inline void getAllY(iterator_rev<T,R> it, T * t){return allY(it,t);}
    inline void getAllY(iterator_step<T,R> it, T * t){return allY(it,t);}
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getLeftUp, getUp, getRightUp, getRight, getRightDown, getDown, getLeftDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get8(iterator<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_cont<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_step<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_rev<T,R> it){return eight(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getUp, getRight, getDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 4 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get4(iterator<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_cont<T,R> it){return four(it._rank,it._width);}
     inline T * get4(iterator_step<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_rev<T,R> it){return four(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain, for physical left border iterators
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInRight(iterator_phb_left<T,R> it){return inright(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain, for physical right border iterators
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInLeft(iterator_phb_right<T,R> it){return inleft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain, for physical down border iterators
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInUp(iterator_phb_down<T,R> it){return inup(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain, for physical up border iterators
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInDown(iterator_phb_up<T,R> it){return indown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    
  };
  
  //================================================================================
  //!  DMatrix_impl class
  /*!
    template of the distributed matrix implementation of SkelGIS. This is the second specialization of the DMatrix_impl.
    \tparam T is the type of data to store in the DMatrix_impl
    R overlap template parameter is not asked and specialized to 0.
    The default value of line is used (true).
  */
  //-------------------------------------------------------------------------------
  template<class T> struct DMatrix_impl<T,0,true> : DMatrix_base<T>
  //-------------------------------------------------------------------------------
  {
  protected:

    //-------------------------------------------------------------------------------
    //! Creation of the ditributed matrix
    /*!
      Data distribution and construction of the parallel data structure.
    */
    //-------------------------------------------------------------------------------
    void create()
    //-------------------------------------------------------------------------------
    {
      if(this->local==false)
	{
	  //division of the matrix in cols and rows
	  this->cols = 1;
	  this->rows = Mpi_::mpi_nb;

	  this->row=Mpi_::mpi_rank;
	  this->col=0;

	  this->loc_head.width=this->head.width;

	  this-> remainderh = 0;
	  //if the height of the matrix can be divided by the number of rows
	  if(this->head.height%this->rows==0)
	    this->loc_head.height=floor(this->head.height/this->rows);
	  //else we add one to the height of the submatrix and the last row will have less elements
	  else
	    {
	      this->remainderh = this->head.height - floor(this->head.height/this->rows)*this->rows;

	      if(this->row<this->remainderh)
		this->loc_head.height=floor(this->head.height/this->rows)+1;
	      else
		this->loc_head.height=floor(this->head.height/this->rows);
	    }

	  this->loc_head.x=this->head.x;
	  this->loc_head.y=this->head.y-this->row*this->loc_head.height*this->head.spacing;
	  this->loc_head.spacing=this->head.spacing;
	  this->loc_head.nodata=this->head.nodata;
	}
      else
	{
	  this->loc_head.x = this->head.x;
	  this->loc_head.y = this->head.y;
	  this->loc_head.width = this->head.width;
	  this->loc_head.height = this->head.height;
	  this->loc_head.spacing = this->head.spacing;
	  this->loc_head.nodata = this->head.nodata;

	  this->col=0;this->row=0;
	  this->cols=1;this->rows=1;
	  this-> remainderh = 0;
	}
      
      this->data = new T[this->loc_head.width*this->loc_head.height];
    }
    //-------------------------------------------------------------------------------
    //! Values initialization of the elements of the DMatrix_impl
    /*!
      \param value is the value to set to the elements
    */
    //-------------------------------------------------------------------------------
    void initValues(T value)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<this->loc_head.width*this->loc_head.height;i++)
	{
	  this->data[i]=value;
	}
    }
    //-------------------------------------------------------------------------------
    //! Read the zone of the binary file concerned for the current MPI process
    /*!
      \param binFile is binary file to read
    */
    //-------------------------------------------------------------------------------
    void read(const char * binFile)
    //-------------------------------------------------------------------------------
    {
      //read the zone of the file for my own block of data
      std::ifstream f(binFile, std::ios::binary | std::ios::in);

      //move to the begining of the bloc of data concerned
      unsigned int mvline;
      if(Mpi_::mpi_rank<this->remainderh)
	mvline = sizeof(HEADER) + (this->row*this->loc_head.height*this->head.width)*sizeof(T);
      else
	mvline = sizeof(HEADER) + ((this->remainderh*(this->loc_head.height+1)+(this->row-this->remainderh)*this->loc_head.height)*this->head.width)*sizeof(T);

      f.seekg(mvline,std::ios::beg);

      //read by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  for(unsigned int k=0;k<this->loc_head.width;k++)
	    {
	      T val;
	      unsigned int rank = j*this->loc_head.width + k;
	      f.read(reinterpret_cast<char*>(&val),sizeof(T));
	      this->data[rank]=val;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the indexes associated to a precise rank
    /*!
      \param rank is the rank at which the indexes values are wanted
      \param col is the column return value
      \param li is the line return value
    */
    //-------------------------------------------------------------------------------
    void indexesIt(unsigned int rank,int &col,int &li)
    //-------------------------------------------------------------------------------
    {
      li = floor(rank/this->loc_head.width);
      col = rank%this->loc_head.width;
    }
    //-------------------------------------------------------------------------------

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix_impl() : DMatrix_base<T>(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      This constructor will simply allocate the good space.
      \param h is the header to use to build your matrix
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(HEADER h,const T value, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head=h;
      this->local=loc;
      create();
      initValues(value);
    }
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      The constructor evaluate how to divide the matrix in submatrice and do the work.
      \param binFile is the path of the binary file you want to work on
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(const char * binFile, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head = Header::read(binFile);
      this->local=loc;
      create();
      initValues(0);
      read(binFile);
    }
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix_impl()
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	delete [] this->data;
    }
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> begin(){return iterator<T,0>(0);}
    inline iterator_cont<T,0> begin_cont(){return iterator_cont<T,0>(0);}
    inline iterator_step<T,0> begin_step(int step,int nb){return iterator_step<T,0>(0,step,nb);}
    inline iterator_rev<T,0> begin_rev(){return iterator_rev<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_line<T,0> begin_line(){return iterator_line<T,0>(0,this->loc_head.width);}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> end(){return iterator<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_cont<T,0> end_cont(){return iterator_cont<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_step<T,0> end_step(){return iterator_step<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_rev<T,0> end_rev(){return iterator_rev<T,0>(0);}
    inline iterator_line<T,0> end_line(){return iterator_line<T,0>(this->loc_head.width*this->loc_head.height-this->loc_head.width,this->loc_head.width);}
    //-------------------------------------------------------------------------------
    //! get the iterator on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator<T,0> getIterator(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank =li*this->loc_head.width+col;
	  return iterator<T,0>(rank,this->loc_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator contiguous on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_cont<T,0> getIterator_cont(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = li*this->loc_head.width+col;
	  return iterator_cont<T,0>(rank,this->loc_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator reverse on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_rev<T,0> getIterator_rev(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = li*this->loc_head.width+col;
	  return iterator_rev<T,0>(rank,this->loc_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the indexes of the iterator on the matrix
    /*!
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    void getIndexes(iterator<T,0> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_cont<T,0> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_rev<T,0> it,int &col,int &li){indexesIt(it._rank,col,li);}
    //-------------------------------------------------------------------------------
    //! Print the matrix by bloc
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      std::stringstream st;
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      st<<this->data[i*this->loc_head.width+j]<<" ";
	    }
	  st<<"\n";
	}
      Mpi_::print(st.str());
    }
    //-------------------------------------------------------------------------------
    //! Write the zone file concerned for the current MPI process
    /*!
      write the MPI prcess zone, and spread the data in the subMatrices OpenMP
      only one submatrice if OpenMP is disabled
      \param binFile is binary file to write in
    */
    //-------------------------------------------------------------------------------
    void write(char * binFile)
    //-------------------------------------------------------------------------------
    {
      MPI_File fh;
      MPI_Status status;
      //open the output file
      MPI_File_open(MPI_COMM_WORLD, binFile,
		MPI_MODE_CREATE|MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);
   
      // process 0 writes the HEADER of the file
      if (Mpi_::mpi_rank==0)
	MPI_File_write_at(fh, 0,reinterpret_cast<char*>(&this->head) ,sizeof(HEADER) , MPI_CHAR, &status);
      //the begining of the bloc of write in
      unsigned int mvline;
      if (Mpi_::mpi_rank<this->remainderh)
	mvline=sizeof(HEADER)+this->row*this->loc_head.height*this->loc_head.width*sizeof(T);
      else
	mvline=sizeof(HEADER)+(this->row*this->loc_head.height+this->remainderh)*this->loc_head.width*sizeof(T);

      // pointer in the data of the dmatrix
      T *data_pointer=this->data;

      //write by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  MPI_File_write_at(fh, mvline,reinterpret_cast<char*>(data_pointer) ,this->loc_head.width*sizeof(T) , MPI_CHAR, &status);
	  data_pointer+=this->loc_head.width;
	  mvline+=this->loc_head.width*sizeof(T);
	}
      MPI_File_close(&fh);
    }
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //!  DMatrix_impl class
  /*!
    template of the distributed matrix implementation of SkelGIS. This is the first specialization of the DMatrix_impl.
    \tparam T is the type of data to store in the DMatrix_impl
    \tparam R is the overlap distance needed by the calculation, in other words, the size of the physical border needed
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct DMatrix_impl<T,R,false> : public DMatrix_base<T>
  //-------------------------------------------------------------------------------
  {
  public:

    HEADER border_head; /*!< Header defining the current local matrix with overlap R */

  protected:
    int mpi_ranks[8]; /*!< table of size 8 for the neighbor processes to make MPI exchanges */
    unsigned int beginPhBLeft,beginPhBRight,beginPhBUp,beginPhBDown; /*!< begining indexes for the physical borders iterators */
    unsigned int endPhBLeft,endPhBRight,endPhBUp,endPhBDown; /*!< ending indexes for the physical borders iterators */
    //-------------------------------------------------------------------------------
    //! Creation of the ditributed matrix
    /*!
      Data distribution and construction of the parallel data structure.
    */
    //-------------------------------------------------------------------------------
    void create()
    //-------------------------------------------------------------------------------
    {
      if(this->local==false)
	{
	  //division of the matrix in cols and rows
	  this->cols = floor(sqrt(Mpi_::mpi_nb));
	  int nb = 1;
	  while(Mpi_::mpi_nb%this->cols!=0)
	    {
	      this->cols = floor(sqrt(Mpi_::mpi_nb))-nb;
	      nb++;
	    }
	  this->rows = Mpi_::mpi_nb/this->cols;

	  this->row=floor(Mpi_::mpi_rank/this->cols);
	  this->col=Mpi_::mpi_rank%this->cols;

	  this-> remainderh = 0;
	  //if the height of the matrix can be divided by the number of rows
	  if(this->head.height%this->rows==0)
	    this->loc_head.height=floor(this->head.height/this->rows);
	  //else we add one to the height of the submatrix and the last row will have less elements
	  else
	    {
	      this->remainderh = this->head.height - floor(this->head.height/this->rows)*this->rows;

	      if(this->row<this->remainderh)
		this->loc_head.height=floor(this->head.height/this->rows)+1;
	      else
		this->loc_head.height=floor(this->head.height/this->rows);
	    }

	  this-> remainderw = 0;
	  //if the height of the matrix can be divided by the number of rows
	  if(this->head.width%this->cols==0)
	    this->loc_head.width=floor(this->head.width/this->cols);
	  //else we add one to the height of the submatrix and the last row will have less elements
	  else
	    {
	      this->remainderw = this->head.width - floor(this->head.width/this->cols)*this->cols;

	      if(this->col<this->remainderw)
		this->loc_head.width=floor(this->head.width/this->cols)+1;
	      else
		this->loc_head.width=floor(this->head.width/this->cols);
	    }
	
	  this->loc_head.x=this->head.x+this->col*this->loc_head.width*this->head.spacing;
	  this->loc_head.y=this->head.y-this->row*this->loc_head.height*this->head.spacing;
	  this->loc_head.spacing=this->head.spacing;
	  this->loc_head.nodata=this->head.nodata;
	}
      else
	{
	  this->loc_head.x = this->head.x;
	  this->loc_head.y = this->head.y;
	  this->loc_head.width = this->head.width;
	  this->loc_head.height = this->head.height;
	  this->loc_head.spacing = this->head.spacing;
	  this->loc_head.nodata = this->head.nodata;

	  this->col=0;this->row=0;
	  this->cols=1;this->rows=1;
	  this-> remainderw = 0;
	  this-> remainderh = 0;
	}

      this->border_head.x = this->loc_head.x - R*this->loc_head.spacing;
      this->border_head.y = this->loc_head.y - R*this->loc_head.spacing;
      this->border_head.width = this->loc_head.width + R*2;
      this->border_head.height = this->loc_head.height + R*2;
      this->data = new T[this->border_head.width*this->border_head.height];
    }
    //-------------------------------------------------------------------------------
    //! Values initialization of the elements of the DMatrix_impl
    /*!
      \param value is the value to set to the elements
    */
    //-------------------------------------------------------------------------------
    void initValues(T value)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<this->border_head.width*this->border_head.height;i++)
	{
	  this->data[i]=value;
	}
    }
    //-------------------------------------------------------------------------------
    //! Calculation of the physical border limits
    //-------------------------------------------------------------------------------
    void phBLimits()
    //-------------------------------------------------------------------------------
    {
      if(this->col==0)
	{
	  beginPhBLeft = R*this->border_head.width;
	  endPhBLeft = this->border_head.width*(this->border_head.height-R-1)+R;
	}
      else
	{
	  beginPhBLeft = 0;
	  endPhBLeft = 0;
	}

      if(this->col==this->cols-1)
	{
	  beginPhBRight = (R+1)*this->border_head.width-R;
	  endPhBRight = (R+1+this->loc_head.height-1)*this->border_head.width;
	}
      else
	{
	  beginPhBRight = 0;
	  endPhBRight = 0;
	}

      if(this->row==0)
	{
	  beginPhBUp = R;
	  endPhBUp = R*this->border_head.width-R;
	}
      else
	{
	  beginPhBUp = 0;
	  endPhBUp = 0;
	}

      if(this->row==this->rows-1)
	{
	  beginPhBDown = this->border_head.width*(this->loc_head.height+R)+R;
	  endPhBDown = this->border_head.width*this->border_head.height-R;
	}
      else
	{
	  beginPhBDown = 0;
	  endPhBDown = 0;
	}
    }
    //-------------------------------------------------------------------------------
    //! Read the zone of the binary file concerned for the current MPI process
    /*!
      \param binFile is binary file to read
    */
    //-------------------------------------------------------------------------------
    void read(const char * binFile)
    //-------------------------------------------------------------------------------
    {
      //read the zone of the file for my own block of data
      std::ifstream f(binFile, std::ios::binary | std::ios::in);

      //move to the begining of the bloc of data concerned
      unsigned int mvline;
      if(this->row<this->remainderh && this->col<this->remainderw)
	mvline = sizeof(HEADER) + (this->col*this->loc_head.width
				   +this->row*this->loc_head.height*this->head.width)*sizeof(T);
      else if(this->row<this->remainderh)
	mvline = sizeof(HEADER) + (this->remainderw*(this->loc_head.width+1)+(this->col-this->remainderw)*this->loc_head.width
				   +(this->row*this->loc_head.height)
				   *this->head.width)*sizeof(T);
      else if(this->col<this->remainderw)
	mvline = sizeof(HEADER) + (this->col*this->loc_head.width
				   +(this->remainderh*(this->loc_head.height+1)+(this->row-this->remainderh)*this->loc_head.height)
				   *this->head.width)*sizeof(T);
      else
	mvline = sizeof(HEADER) + (this->remainderw*(this->loc_head.width+1)+(this->col-this->remainderw)*this->loc_head.width
				   +(this->remainderh*(this->loc_head.height+1)+(this->row-this->remainderh)*this->loc_head.height)
				   *this->head.width)*sizeof(T);

      f.seekg(mvline,std::ios::beg);

      //read by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  if(j>0)
	    {
	      //move to the next line of the bloc
	      unsigned int mv = this->head.width - this->loc_head.width;
	      f.seekg(mv*sizeof(T),std::ios::cur);
	    }

	  for(unsigned int k=0;k<this->loc_head.width;k++)
	    {
	      T val;
	      unsigned int rank = R*this->border_head.width + j*this->border_head.width + k + R;
	      f.read(reinterpret_cast<char*>(&val),sizeof(T));
	      this->data[rank]=val;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! Get rank from a column number and a row number
    /*!
      \param column is the column number
      \param columns is the total number of columns
      \param row is the row number
      \return the rank
    */
    //-------------------------------------------------------------------------------
    unsigned int getRank(int c, int cs, int r)
    //-------------------------------------------------------------------------------
    {
      unsigned int res = r*cs+c;
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get ranks
    /*!
      Get the ranks of the 8 neighbors for MPI exchanges
      \return 
      if the value is -55555 then there is no neighbor in this direction
      if the value is >0 it represent the MPI rank concerned by the exchange
    */
    //-------------------------------------------------------------------------------
    void getNeighborRanks()
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<8;i++)
	mpi_ranks[i] = -55555;
      //up
      if((this->row-1)>=0)
	mpi_ranks[1] = getRank(this->col,this->cols,(this->row-1));
      //down
      if((this->row+1)<this->rows)
	mpi_ranks[5] = getRank(this->col,this->cols,(this->row+1));
      //right
      if((this->col+1)<this->cols)
	mpi_ranks[3] = getRank((this->col+1),this->cols,this->row);
      //left
      if((this->col-1)>=0)
	mpi_ranks[7] = getRank((this->col-1),this->cols,this->row);
      //up left
      if((this->col-1)>=0 && (this->row-1)>=0)
	mpi_ranks[0] = getRank((this->col-1),this->cols,(this->row-1));
      //up right
      if((this->col+1)<this->cols && (this->row-1)>=0)
	mpi_ranks[2] = getRank((this->col+1),this->cols,(this->row-1));
      //down left
      if((this->col-1)>=0 && (this->row+1)<this->rows)
	mpi_ranks[6] = getRank((this->col-1),this->cols,(this->row+1));
      //down right
      if((this->col+1)<this->cols && (this->row+1)<this->rows)
	mpi_ranks[4] = getRank((this->col+1),this->cols,(this->row+1));
    }
    //-------------------------------------------------------------------------------
    //! Get the right border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getRightBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[this->loc_head.height*R];
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(int j=0;j<R;j++)
	    {
	      unsigned int r1 = i*R+j;
	      unsigned int r2 = ((R+1)*border_head.width - R*2) + i*border_head.width +j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the right border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setRightBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(int j=0;j<R;j++)
	    this->data[(border_head.width*R + border_head.width - R) + i*border_head.width + j] = values[i*R+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the left border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getLeftBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[this->loc_head.height*R];
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(int j=0;j<R;j++)
	    {
	      unsigned int r1 = i*R+j;
	      unsigned int r2 = (R*border_head.width + R) + i*border_head.width + j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the left border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setLeftBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(int j=0;j<R;j++)
	    this->data[(border_head.width*R) + i*border_head.width + j] = values[i*R+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the up border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getUpBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[this->loc_head.width*R];
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      unsigned int r1 = i*this->loc_head.width+j;
	      unsigned int r2 = (R*border_head.width + R) + i*border_head.width+j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the up border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setUpBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    this->data[(R) + i*border_head.width+j] = values[i*this->loc_head.width+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the down border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getDownBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[this->loc_head.width*R];
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      unsigned int r1 = i*this->loc_head.width+j;
	      unsigned int r2 = (border_head.height*border_head.width - 2*R*border_head.width + R) + i*border_head.width + j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the down border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setDownBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    this->data[(border_head.width*border_head.height - border_head.width*R + R) + i*border_head.width +j] = values[i*this->loc_head.width+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the down right border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getDownRightBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[R*R];
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    {
	      unsigned int r1 = i*R+j;
	      unsigned int r2 = (border_head.height*border_head.width - (2*R-1)*border_head.width - R*2) + i*border_head.width +j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the down right border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setDownRightBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    this->data[(border_head.width*border_head.height - (R-1)*border_head.width - R) + i*border_head.width + j] = values[i*R+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the down left border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getDownLeftBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[R*R];
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    {
	      unsigned int r1 = i*R+j;
	      unsigned int r2 = (border_head.height*border_head.width - 2*R*border_head.width + R) + i*border_head.width + j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the down left border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setDownLeftBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    this->data[(border_head.width*border_head.height - border_head.width*R) + i*border_head.width +j] = values[i*R+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the up left border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getUpLeftBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[R*R];
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    {
	      unsigned int r1 = i*R+j;
	      unsigned int r2 = (border_head.width*R + R) + i*border_head.width + j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the up left border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setUpLeftBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    this->data[i*border_head.width+j] = values[i*R+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the up right border to send to neighbor processors
    /*!
      \return list of values
    */
    //-------------------------------------------------------------------------------
    T * getUpRightBorderToSend()
    //-------------------------------------------------------------------------------
    {
      T * result = new T[R*R];
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    {
	      unsigned int r1 = i*R+j;
	      unsigned int r2 = ((R+1)*border_head.width - R*2) + i*border_head.width +j;
	      result[r1] = this->data[r2];
	    }
	}
      return result;
    }
    //-------------------------------------------------------------------------------
    //! Set the up right border ghost cells with neighbor values from processors
    /*!
      \param values is the table values to fill ghost cells with
    */
    //-------------------------------------------------------------------------------
    void setUpRightBorder(T * values)
    //-------------------------------------------------------------------------------
    {
      for(int i=0;i<R;i++)
	{
	  for(int j=0;j<R;j++)
	    this->data[(border_head.width-R) + i*border_head.width+j] = values[i*R+j];
	}
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRight(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank+1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank-1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r--;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r - width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r + width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width -1 ;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r + width -1 ;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r - width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width - 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[r];
	  r = r - width - 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T right(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + 1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T left(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - 1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T up(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width;
      return this->data[r];
    }//-------------------------------------------------------------------------------
    //! Get the nearest neighbor down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T down(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T rightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width +1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T leftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width -1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T rightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width +1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T leftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width -1;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allX(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[b];
	  b++;
	}
      b++;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = this->data[b];
	  b++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allX(unsigned int rank, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
        {
          t[i] = this->data[b];
          b++;
        }
      b++;
      for(int i=R;i<2*R;i++)
        {
          t[i] = this->data[b];
          b++;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allY(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
	{
	  res[i] = this->data[b];
	  b = b + width;
	}
      b = b + width;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = this->data[b];
	  b = b + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allY(unsigned int rank,unsigned int width, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
        {
          t[i] = this->data[b];
          b = b + width;
        }
      b = b + width;
      for(int i=R;i<2*R;i++)
        {
          t[i] = this->data[b];
          b = b + width;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to leftUp, up, rightUp, right, rightDown, down, leftDown and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * eight(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = leftUp(rank,width);
      res[1] = up(rank,width);
      res[2] = rightUp(rank,width);
      res[3] = right(rank);
      res[4] = rightDown(rank,width);
      res[5] = down(rank,width);
      res[6] = leftDown(rank,width);
      res[7] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to up, right, down and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * four(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = up(rank,width);
      res[1] = right(rank);
      res[2] = down(rank,width);
      res[3] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T inright(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = rank%border_head.width;
      unsigned int r = rank + (R-re);
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T inleft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = border_head.width - rank%border_head.width -1 ;
      unsigned int r = rank - (R-re);
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T inup(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = border_head.height-floor(rank/border_head.width)-1;
      unsigned int r = rank - (R-re)*width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T indown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int re = floor(rank/border_head.width);
      unsigned int r = rank + (R-re)*width;
      return this->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the indexes associated to a precise rank
    /*!
      \param rank is the rank at which the indexes values are wanted
      \param col is the column return value
      \param li is the line return value
    */
    //-------------------------------------------------------------------------------
    void indexesIt(unsigned int rank,int &col,int &li)
    //-------------------------------------------------------------------------------
    {
      unsigned int relative_rank = rank - R*border_head.width - (floor(rank/border_head.width)-R)*R*2 -R;
      li = floor(relative_rank/this->loc_head.width);
      col = relative_rank%this->loc_head.width;
    }
    //-------------------------------------------------------------------------------

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix_impl() : DMatrix_base<T>(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      This constructor will simply allocate the good space.
      \param h is the header to use to build your matrix
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(HEADER h,const T value, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head=h;
      this->local=loc;
      create();
      initValues(value);
      getNeighborRanks();
      phBLimits();
    }
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      The constructor evaluate how to divide the matrix in submatrice and do the work.
      \param binFile is the path of the binary file you want to work on
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(const char * binFile, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head = Header::read(binFile);
      this->local=loc;
      create();
      initValues(0);
      read(binFile);
      getNeighborRanks();
      phBLimits();
    }
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix_impl()
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	delete [] this->data;
    }
    //-------------------------------------------------------------------------------
    //! Set value in the global middle
    /*!
      \param val is the value to affect
    */
    //-------------------------------------------------------------------------------
    void setGlobalMiddleValue(T val)
    //-------------------------------------------------------------------------------
    {
      if((this->col!=this->cols-1 && this->row!=this->rows-1) ||
	 (this->cols==1 && this->row!=this->rows-1) ||
	 (this->col!=this->cols-1 && this->rows==1))
	{
	  unsigned int i = floor(this->head.height/2)-1;
	  unsigned int j = floor(this->head.width/2)-1;
	  if(this->col*this->loc_head.width<=j && 
	     (this->col+1)*this->loc_head.width>j && 
	     this->row*this->loc_head.height<=i && 
	     (this->row+1)*this->loc_head.height>i)
	    {
	      unsigned int j_loc = j-this->col*this->loc_head.width;
	      unsigned int i_loc = i-this->row*this->loc_head.height;
	      unsigned int rank = R*border_head.width +  i_loc*border_head.width + j_loc + R;
	      this->data[rank] = val;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! to set all the values of the physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setPhysicalBorder(T val)
    {
      iterator_phb_left<T,R> left =begin_phb_left();
      iterator_phb_left<T,R> leftEnd = end_phb_left();
      for(left;left<leftEnd;++left)
	this->data[left._rank] = val;
      iterator_phb_right<T,R> right =begin_phb_right();
      iterator_phb_right<T,R> rightEnd = end_phb_right();
      for(right;right<rightEnd;++right)
	this->data[right._rank] = val;
      iterator_phb_up<T,R> up =begin_phb_up();
      iterator_phb_up<T,R> upEnd = end_phb_up();
      for(up;up<upEnd;++up)
	this->data[up._rank] = val;
      iterator_phb_down<T,R> down =begin_phb_down();
      iterator_phb_down<T,R> downEnd = end_phb_down();
      for(down;down<downEnd;++down)
	this->data[down._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the right of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setRightPhysicalBorder(T val)
    {
      iterator_phb_right<T,R> it =begin_phb_right();
      iterator_phb_right<T,R> itEnd = end_phb_right();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the left of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setLeftPhysicalBorder(T val)
    {
      iterator_phb_left<T,R> it =begin_phb_left();
      iterator_phb_left<T,R> itEnd = end_phb_left();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the up of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setUpPhysicalBorder(T val)
    {
      iterator_phb_up<T,R> it =begin_phb_up();
      iterator_phb_up<T,R> itEnd = end_phb_up();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! to set all the values on the down of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    void setDownPhysicalBorder(T val)
    {
      iterator_phb_down<T,R> it =begin_phb_down();
      iterator_phb_down<T,R> itEnd = end_phb_down();
      for(it;it<itEnd;++it)
	this->data[it._rank] = val;
    }
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> begin(){return iterator<T,R>(R*this->border_head.width+R,this->border_head.width);}
    inline iterator_cont<T,R> begin_cont(){return iterator_cont<T,R>(R*this->border_head.width+R,this->border_head.width);}
    inline iterator_rev<T,R> begin_rev(){return iterator_rev<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_step<T,R> begin_step(int step,int nb){return iterator_step<T,R>(R*this->border_head.width+R,this->border_head.width,step,nb);}
    inline iterator_phb_left<T,R> begin_phb_left(){return iterator_phb_left<T,R>(beginPhBLeft,this->border_head.width);}
    inline iterator_phb_right<T,R> begin_phb_right(){return iterator_phb_right<T,R>(beginPhBRight,this->border_head.width);}
    inline iterator_phb_up<T,R> begin_phb_up(){return iterator_phb_up<T,R>(beginPhBUp,this->border_head.width);}
    inline iterator_phb_down<T,R> begin_phb_down(){return iterator_phb_down<T,R>(beginPhBDown,this->border_head.width);}
    inline iterator_line<T,R> begin_line(){return iterator_line<T,R>(R*this->border_head.width+R,this->border_head.width);}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> end(){return iterator<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_cont<T,R> end_cont(){return iterator_cont<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_rev<T,R> end_rev(){return iterator_rev<T,R>(R*this->border_head.width+R-1,this->border_head.width);}
    inline iterator_step<T,R> end_step(){return iterator_step<T,R>((this->border_head.width*this->border_head.height)-1-R*this->border_head.width-R,this->border_head.width);}
    inline iterator_phb_left<T,R> end_phb_left(){return iterator_phb_left<T,R>(endPhBLeft,this->border_head.width);}
    inline iterator_phb_right<T,R> end_phb_right(){return iterator_phb_right<T,R>(endPhBRight,this->border_head.width);}
    inline iterator_phb_up<T,R> end_phb_up(){return iterator_phb_up<T,R>(endPhBUp,this->border_head.width);}
    inline iterator_phb_down<T,R> end_phb_down(){return iterator_phb_down<T,R>(endPhBDown,this->border_head.width);}
    inline iterator_line<T,R> end_line(){return iterator_line<T,R>((this->border_head.width*this->border_head.height)-(R+1)*this->border_head.width+R,this->border_head.width);}
    //-------------------------------------------------------------------------------
    //! get the iterator on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator<T,R> getIterator(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = R*border_head.width + R + li*border_head.width+col;
	  return iterator<T,R>(rank,this->border_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator contiguous on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_cont<T,R> getIterator_cont(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = R*border_head.width + R + li*border_head.width+col;
	  return iterator_cont<T,R>(rank,this->border_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator reverse on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_rev<T,R> getIterator_rev(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = R*border_head.width + R + li*border_head.width+col;
	  return iterator_rev<T,R>(rank,this->border_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the indexes of the iterator on the matrix
    /*!
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    void getIndexes(iterator<T,R> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_cont<T,R> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_rev<T,R> it,int &col,int &li){indexesIt(it._rank,col,li);}
    //-------------------------------------------------------------------------------
    //! Print the matrix by bloc
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      std::stringstream st;
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      st<<this->data[R*this->border_head.width + R + i*this->border_head.width+j]<<" ";
	    }
	  st<<"\n";
	}
      Mpi_::print(st.str());
    }
    //-------------------------------------------------------------------------------
    //! Print all the matrix by bloc
    //-------------------------------------------------------------------------------
    void printAll()
    //-------------------------------------------------------------------------------
    {
      std::stringstream st;
      for(unsigned int i=0;i<this->border_head.height;i++)
        {
          for(unsigned int j=0;j<this->border_head.width;j++)
            {
              st<<this->data[i*this->border_head.width+j]<<" ";
            }
          st<<"\n";
        }
      Mpi_::print(st.str());
    }
    //-------------------------------------------------------------------------------
    //! Write the zone file concerned for the current MPI process
    /*!
      write the MPI prcess zone, and spread the data in the subMatrices OpenMP
      only one submatrice if OpenMP is disabled
      \param binFile is binary file to write in
    */
    //-------------------------------------------------------------------------------
    void write(char * binFile)
    //-------------------------------------------------------------------------------
    {
      MPI_File fh;
      MPI_Status status;
      //open the output file
      MPI_File_open(MPI_COMM_WORLD, binFile,
		MPI_MODE_CREATE|MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);
   
      // process 0 writes the HEADER of the file
      if (Mpi_::mpi_rank==0)
	MPI_File_write_at(fh, 0,reinterpret_cast<char*>(&this->head) ,sizeof(HEADER) , MPI_CHAR, &status);
      //the begining of the bloc of write in
      unsigned int mvline;
      // number of bytes between 2 blocks
      unsigned int step;
      
      // the starting point of the row where the DMatrix is saved
      if (this->row<this->remainderh)
	mvline=sizeof(HEADER)+sizeof(T)*this->head.width*this->loc_head.height*this->row;
      else
	mvline=sizeof(HEADER)+sizeof(T)*this->head.width*(this->loc_head.height*this->row+this->remainderh);
      
      // now we must add the number of bytes corresponding to the column of the processor
      if (this->col<this->remainderw)
	mvline+=sizeof(T)*this->loc_head.width*this->col;
      else
	mvline+=sizeof(T)*(this->loc_head.width*this->col+this->remainderw);

      // pointer in the data of the dmatrix
      T *data_pointer=this->data+R*(this->border_head.width+1);

      //write by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  std::stringstream st1;
	  st1<<"mvline :"<<mvline<<" loc head width "<<this->loc_head.width<<" loc head height "
	     <<this->loc_head.height << " head width "<<this->head.width<<" head height "<<this->head.height
	     <<" row "<<this->row<<" col "<<this->col<<"\n";
	  Mpi_::print(st1.str());
	  MPI_File_write_at(fh, mvline,reinterpret_cast<char*>(data_pointer) ,this->loc_head.width*sizeof(T) , MPI_CHAR, &status);
	  data_pointer+=this->border_head.width;
	  mvline+=this->head.width*sizeof(T);
	}
      MPI_File_close(&fh);
    }
    //-------------------------------------------------------------------------------
    //! get borders
    /*!
      Get the borders from other processors
      MPI exchanges
    */
    //-------------------------------------------------------------------------------
    void getBorders()
    //-------------------------------------------------------------------------------
    {	  
      if(this->col!=0 && this->row!=0)
	{
	  T * toSend = getUpLeftBorderToSend();
	  T * toGet = new T[R*R];
	  Communications<T>::Exchanges(toSend,toGet,R*R,mpi_ranks[0]);
	  setUpLeftBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
      if(this->col!=(this->cols-1) && this->row!=(this->rows-1))
	{
	  T * toSend = getDownRightBorderToSend();
	  T * toGet = new T[R*R];
	  Communications<T>::Exchanges(toSend,toGet,R*R,mpi_ranks[4]);
	  setDownRightBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
      if(this->row!=0)
	{
	  T * toSend = getUpBorderToSend();
	  T * toGet = new T[this->loc_head.width*R];
	  Communications<T>::Exchanges(toSend,toGet,this->loc_head.width*R,mpi_ranks[1]);
	  setUpBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
      if(this->row!=(this->rows-1))
	{     
	  T * toSend = getDownBorderToSend();
	  T * toGet = new T[this->loc_head.width*R];
	  Communications<T>::Exchanges(toSend,toGet,this->loc_head.width*R,mpi_ranks[5]);
	  setDownBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
      if(this->col!=(this->cols-1) && this->row!=0)
	{
	  T * toSend = getUpRightBorderToSend();
	  T * toGet = new T[R*R];
	  Communications<T>::Exchanges(toSend,toGet,R*R,mpi_ranks[2]);
	  setUpRightBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
      if(this->row!=(this->rows-1) && this->col!=0)
	{
	  T * toSend = getDownLeftBorderToSend();
	  T * toGet = new T[R*R];
	  Communications<T>::Exchanges(toSend,toGet,R*R,mpi_ranks[6]);
	  setDownLeftBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
	  
      if(this->col!=(this->cols-1))
	{
	  T * toSend = getRightBorderToSend();
	  T * toGet = new T[this->loc_head.height*R];
	  Communications<T>::Exchanges(toSend,toGet,this->loc_head.height*R,mpi_ranks[3]);
	  setRightBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
      if(this->col!=0)
	{
	  T * toSend = getLeftBorderToSend();
	  T * toGet = new T[this->loc_head.height*R];
	  Communications<T>::Exchanges(toSend,toGet,this->loc_head.height*R,mpi_ranks[7]);
	  setLeftBorder(toGet);
	  delete [] toSend;
	  delete [] toGet;
	}
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right of the current it.
      This method exists for each type of iterator with right values in SkelGIS.
      \param it is the iterator position
      \return a table of right values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRight(iterator<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_cont<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_step<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_rev<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_phb_left<T,R> it){return allRight(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left of the current it.
      This method exists for each type of iterator with left values in SkelGIS.
      \param it is the iterator position
      \return a table of left values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeft(iterator<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_cont<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_step<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_rev<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_phb_right<T,R> it){return allLeft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the up of the current it.
      This method exists for each type of iterator with up values in SkelGIS.
      \param it is the iterator position
      \return a table of up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllUp(iterator<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_cont<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_step<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_rev<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_phb_down<T,R> it){return allUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the down of the current it.
      This method exists for each type of iterator with down values in SkelGIS.
      \param it is the iterator position
      \return a table of down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllDown(iterator<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_cont<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_step<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_rev<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_phb_up<T,R> it){return allDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right down of the current it.
      This method exists for each type of iterator with right down values in SkelGIS.
      \param it is the iterator position
      \return a table of right down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightDown(iterator<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_cont<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_step<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_rev<T,R> it){return allRightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left down of the current it.
      This method exists for each type of iterator with left down values in SkelGIS.
      \param it is the iterator position
      \return a table of left down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftDown(iterator<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_cont<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_step<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_rev<T,R> it){return allLeftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right up of the current it.
      This method exists for each type of iterator with right up values in SkelGIS.
      \param it is the iterator position
      \return a table of right up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightUp(iterator<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_cont<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_step<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_rev<T,R> it){return allRightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left up of the current it.
      This method exists for each type of iterator with left up values in SkelGIS.
      \param it is the iterator position
      \return a table of left up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftUp(iterator<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_cont<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_step<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_rev<T,R> it){return allLeftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right value in SkelGIS.
      \param it is the iterator position
      \return the right neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRight(iterator<T,R> it){return right(it._rank);}
    inline T getRight(iterator_cont<T,R> it){return right(it._rank);}
    inline T getRight(iterator_step<T,R> it){return right(it._rank);}
    inline T getRight(iterator_rev<T,R> it){return right(it._rank);}
    inline T getRight(iterator_phb_left<T,R> it){return right(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left value in SkelGIS.
      \param it is the iterator position
      \return the left neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeft(iterator<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_cont<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_step<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_rev<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_phb_right<T,R> it){return left(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with up value in SkelGIS.
      \param it is the iterator position
      \return the up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getUp(iterator<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_cont<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_step<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_rev<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_phb_down<T,R> it){return up(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with down value in SkelGIS.
      \param it is the iterator position
      \return the down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getDown(iterator<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_cont<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_step<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_rev<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_phb_up<T,R> it){return down(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right down value in SkelGIS.
      \param it is the iterator position
      \return the right down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightDown(iterator<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_cont<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_step<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_rev<T,R> it){return rightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left down value in SkelGIS.
      \param it is the iterator position
      \return the left down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftDown(iterator<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_cont<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_step<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_rev<T,R> it){return leftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right up value in SkelGIS.
      \param it is the iterator position
      \return the right up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightUp(iterator<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_cont<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_step<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_rev<T,R> it){return rightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left up value in SkelGIS.
      \param it is the iterator position
      \return the left up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftUp(iterator<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_cont<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_step<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_rev<T,R> it){return leftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline T * getAllX(iterator<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_cont<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_step<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_rev<T,R> it){return allX(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline void getAllX(iterator<T,R> it, T * t){return allX(it,t);}
    inline void getAllX(iterator_cont<T,R> it, T * t){return allX(it,t);}
    inline void getAllX(iterator_rev<T,R> it, T * t){return allX(it,t);}
    inline void getAllX(iterator_step<T,R> it, T * t){return allX(it,t);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline T * getAllY(iterator<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_cont<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_step<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_rev<T,R> it){return allY(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline void getAllY(iterator<T,R> it, T * t){return allY(it,t);}
    inline void getAllY(iterator_cont<T,R> it, T * t){return allY(it,t);}
    inline void getAllY(iterator_rev<T,R> it, T * t){return allY(it,t);}
    inline void getAllY(iterator_step<T,R> it, T * t){return allY(it,t);}
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getLeftUp, getUp, getRightUp, getRight, getRightDown, getDown, getLeftDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get8(iterator<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_cont<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_step<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_rev<T,R> it){return eight(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getUp, getRight, getDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 4 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get4(iterator<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_cont<T,R> it){return four(it._rank,it._width);}
     inline T * get4(iterator_step<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_rev<T,R> it){return four(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain, for physical left border iterators
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInRight(iterator_phb_left<T,R> it){return inright(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain, for physical right border iterators
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInLeft(iterator_phb_right<T,R> it){return inleft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain, for physical down border iterators
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInUp(iterator_phb_down<T,R> it){return inup(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain, for physical up border iterators
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInDown(iterator_phb_up<T,R> it){return indown(it._rank,it._width);}
    //-------------------------------------------------------------------------------

  };

  //================================================================================
  //!  DMatrix_impl class
  /*!
    template of the distributed matrix implementation of SkelGIS. This is the first specialization of the DMatrix_impl.
    \tparam T is the type of data to store in the DMatrix_impl
    R overlap template parameter is not asked and specialized to 0.
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template<class T> struct DMatrix_impl<T,0,false> : DMatrix_base<T>
  //-------------------------------------------------------------------------------
  {
  protected:

    //-------------------------------------------------------------------------------
    //! Creation of the ditributed matrix
    /*!
      Data distribution and construction of the parallel data structure.
    */
    //-------------------------------------------------------------------------------
    void create()
    //-------------------------------------------------------------------------------
    {
      if(this->local==false)
	{
	  //division of the matrix in cols and rows
	  this->cols = floor(sqrt(Mpi_::mpi_nb));
	  int nb = 1;
	  while(Mpi_::mpi_nb%this->cols!=0)
	    {
	      this->cols = floor(sqrt(Mpi_::mpi_nb))-nb;
	      nb++;
	    }
	  this->rows = Mpi_::mpi_nb/this->cols;

	  this->row=floor(Mpi_::mpi_rank/this->cols);
	  this->col=Mpi_::mpi_rank%this->cols;

	  this-> remainderh = 0;
	  //if the height of the matrix can be divided by the number of rows
	  if(this->head.height%this->rows==0)
	    this->loc_head.height=floor(this->head.height/this->rows);
	  //else we add one to the height of the submatrix and the last row will have less elements
	  else
	    {
	      this->remainderh = this->head.height - floor(this->head.height/this->rows)*this->rows;

	      if(this->row<this->remainderh)
		this->loc_head.height=floor(this->head.height/this->rows)+1;
	      else
		this->loc_head.height=floor(this->head.height/this->rows);
	    }

	  this-> remainderw = 0;
	  //if the height of the matrix can be divided by the number of rows
	  if(this->head.width%this->cols==0)
	    this->loc_head.width=floor(this->head.width/this->cols);
	  //else we add one to the height of the submatrix and the last row will have less elements
	  else
	    {
	      this->remainderw = this->head.width - floor(this->head.width/this->cols)*this->cols;

	      if(this->col<this->remainderw)
		this->loc_head.width=floor(this->head.width/this->cols)+1;
	      else
		this->loc_head.width=floor(this->head.width/this->cols);
	    }
	
	  this->loc_head.x=this->head.x+this->col*this->loc_head.width*this->head.spacing;
	  this->loc_head.y=this->head.y-this->row*this->loc_head.height*this->head.spacing;
	  this->loc_head.spacing=this->head.spacing;
	  this->loc_head.nodata=this->head.nodata;
	}
      else
	{
	  this->loc_head.x = this->head.x;
	  this->loc_head.y = this->head.y;
	  this->loc_head.width = this->head.width;
	  this->loc_head.height = this->head.height;
	  this->loc_head.spacing = this->head.spacing;
	  this->loc_head.nodata = this->head.nodata;

	  this->col=0;this->row=0;
	  this->cols=1;this->rows=1;
	  this-> remainderw = 0;
	  this-> remainderh = 0;
	}
      
      this->data = new T[this->loc_head.width*this->loc_head.height];
    }
    //-------------------------------------------------------------------------------
    //! Values initialization of the elements of the DMatrix_impl
    /*!
      \param value is the value to set to the elements
    */
    //-------------------------------------------------------------------------------
    void initValues(T value)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<this->loc_head.width*this->loc_head.height;i++)
	{
	  this->data[i]=value;
	}
    }
    //-------------------------------------------------------------------------------
    //! Read the zone file concerned for the current MPI process
    /*!
      read the MPI prcess zone, and spread the data in the subMatrices OpenMP
      only one submatrice if OpenMP is disabled
      \param binFile is binary file to read
    */
    //-------------------------------------------------------------------------------
    void read(const char * binFile)
    //-------------------------------------------------------------------------------
    {
      //read the zone of the file for my own block of data
      std::ifstream f(binFile, std::ios::binary | std::ios::in);

      //move to the begining of the bloc of data concerned
      unsigned int mvline;
      if(this->row<this->remainderh && this->col<this->remainderw)
	mvline = sizeof(HEADER) + (this->col*this->loc_head.width
				   +this->row*this->loc_head.height*this->head.width)*sizeof(T);
      else if(this->row<this->remainderh)
	mvline = sizeof(HEADER) + (this->remainderw*(this->loc_head.width+1)+(this->col-this->remainderw)*this->loc_head.width
				   +(this->row*this->loc_head.height)
				   *this->head.width)*sizeof(T);
      else if(this->col<this->remainderw)
	mvline = sizeof(HEADER) + (this->col*this->loc_head.width
				   +(this->remainderh*(this->loc_head.height+1)+(this->row-this->remainderh)*this->loc_head.height)
				   *this->head.width)*sizeof(T);
      else
	mvline = sizeof(HEADER) + (this->remainderw*(this->loc_head.width+1)+(this->col-this->remainderw)*this->loc_head.width
				   +(this->remainderh*(this->loc_head.height+1)+(this->row-this->remainderh)*this->loc_head.height)
				   *this->head.width)*sizeof(T);

      f.seekg(mvline,std::ios::beg);

      //read by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  if(j>0)
	    {
	      //move to the next line of the bloc
	      unsigned int mv = this->head.width - this->loc_head.width;
	      f.seekg(mv*sizeof(T),std::ios::cur);
	    }

	  for(unsigned int k=0;k<this->loc_head.width;k++)
	    {
	      T val;
	      unsigned int rank = j*this->loc_head.width + k;
	      f.read(reinterpret_cast<char*>(&val),sizeof(T));
	      this->data[rank]=val;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! Get the indexes associated to a precise rank
    /*!
      \param rank is the rank at which the indexes values are wanted
      \param col is the column return value
      \param li is the line return value
    */
    //-------------------------------------------------------------------------------
    void indexesIt(unsigned int rank,int &col,int &li)
    //-------------------------------------------------------------------------------
    {
      li = floor(rank/this->loc_head.width);
      col = rank%this->loc_head.width;
    }
    //-------------------------------------------------------------------------------

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix_impl() : DMatrix_base<T>(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix
    /*!
      This constructor will simply allocate the good space.
      \param h is the header to use to build your matrix
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(HEADER h,const T value, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head=h;
      this->local=loc;
      create();
      initValues(value);
    }
    //! constructor of the distributed matrix
    /*!
      The constructor evaluate how to divide the matrix in submatrice and do the work.
      \param binFile is the path of the binary file you want to work on
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix_impl(const char * binFile, bool loc=false) : DMatrix_base<T>()
    //-------------------------------------------------------------------------------
    {
      this->head = Header::read(binFile);
      this->local=loc;
      create();
      initValues(0);
      read(binFile);
    }
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix_impl()
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	delete [] this->data;
    }
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> begin(){return iterator<T,0>(0);}
    inline iterator_cont<T,0> begin_cont(){return iterator_cont<T,0>(0);}
    inline iterator_step<T,0> begin_step(int step,int nb){return iterator_step<T,0>(0,step,nb);}
    inline iterator_rev<T,0> begin_rev(){return iterator_rev<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_line<T,0> begin_line(){return iterator_line<T,0>(0,this->loc_head.width);}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> end(){return iterator<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_cont<T,0> end_cont(){return iterator_cont<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_step<T,0> end_step(){return iterator_step<T,0>(this->loc_head.width*this->loc_head.height-1);}
    inline iterator_rev<T,0> end_rev(){return iterator_rev<T,0>(0);}
    inline iterator_line<T,0> end_line(){return iterator_line<T,0>(this->loc_head.width*this->loc_head.height-this->loc_head.width,this->loc_head.width);}
    //-------------------------------------------------------------------------------
    //! get the iterator on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator<T,0> getIterator(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = li*this->loc_head.width+col;
	  return iterator<T,0>(rank,this->loc_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator contiguous on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_cont<T,0> getIterator_cont(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = li*this->loc_head.width+col;
	  return iterator_cont<T,0>(rank,this->loc_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the iterator reverse on the matrix at position (x,y)
    /*!
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    iterator_rev<T,0> getIterator_rev(int col,int li)
    //-------------------------------------------------------------------------------
    {
      if(NULL!=this->data)
	{
	  unsigned int rank = li*this->loc_head.width+col;
	  return iterator_rev<T,0>(rank,this->loc_head.width);
	}
    }
    //-------------------------------------------------------------------------------
    //! get the indexes of the iterator on the matrix
    /*!
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    void getIndexes(iterator<T,0> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_cont<T,0> it,int &col,int &li){indexesIt(it._rank,col,li);}
    void getIndexes(iterator_rev<T,0> it,int &col,int &li){indexesIt(it._rank,col,li);}
    //-------------------------------------------------------------------------------
    //! Print the matrix by bloc
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      std::stringstream st;
      for(unsigned int i=0;i<this->loc_head.height;i++)
	{
	  for(unsigned int j=0;j<this->loc_head.width;j++)
	    {
	      st<<this->data[this->loc_head.width + i*this->loc_head.width+j]<<" ";
	    }
	  st<<"\n";
	}
      Mpi_::print(st.str());
    }
    //-------------------------------------------------------------------------------
    //! Write the zone file concerned for the current MPI process
    /*!
      write the MPI prcess zone, and spread the data in the subMatrices OpenMP
      only one submatrice if OpenMP is disabled
      \param binFile is binary file to write in
    */
    //-------------------------------------------------------------------------------
    void write(char * binFile)
    //-------------------------------------------------------------------------------
    {
      MPI_File fh;
      MPI_Status status;
      //open the output file
      MPI_File_open(MPI_COMM_WORLD, binFile,
		MPI_MODE_CREATE|MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);
   
      // process 0 writes the HEADER of the file
      if (Mpi_::mpi_rank==0)
	MPI_File_write_at(fh, 0,reinterpret_cast<char*>(&this->head) ,sizeof(HEADER) , MPI_CHAR, &status);
      //the begining of the bloc of write in
      unsigned int mvline;
      // number of bytes between 2 blocks
      unsigned int step;

      
      // the starting point of the row where the DMatrix is saved
      if (this->row<this->remainderh)
	mvline=sizeof(HEADER)+sizeof(T)*this->head.width*this->loc_head.height*this->row;
      else
	mvline=sizeof(HEADER)+sizeof(T)*this->head.width*(this->loc_head.height*this->row+this->remainderh);
      
      // now we must add the number of bytes corresponding to the column of the processor
      if (this->col<this->remainderw)
	mvline+=sizeof(T)*this->loc_head.width*this->col;
      else
	mvline+=sizeof(T)*(this->loc_head.width*this->col+this->remainderw);

      // pointer in the data of the dmatrix
      T *data_pointer=this->data;

      //write by lines
      for(unsigned int j=0;j<this->loc_head.height;j++)
	{
	  MPI_File_write_at(fh, mvline,reinterpret_cast<char*>(data_pointer) ,this->loc_head.width*sizeof(T) , MPI_CHAR, &status);
	  data_pointer+=this->loc_head.width;
	  mvline+=this->head.width*sizeof(T);
	}
      MPI_File_close(&fh);
    }
    //-------------------------------------------------------------------------------
    
  };
  //-------------------------------------------------------------------------------

}

#endif
