/*! \file iterator_cont.hpp
 *  \brief Contiguous iterator object definition. Two template specializations. This iterator has to be used for contiguous moves in memory, element by element in the DMatrix. Users do not have to construct an iterator object directly (these objects are constructed in the DMatrix) however operators and methods of the iterator are used by the user.
 */
#ifndef ITERATOR_CONT_H
#define ITERATOR_CONT_H

namespace skelgis{

  //================================================================================
  //!  Iterator contiguous class
  /*!
    template of the contiguous iterator of SkelGIS. This is the first specialization of iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_cont
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_cont():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_cont(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_cont(const iterator_cont<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_cont(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_cont<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if((_rank+1)%_width<(_width-R))
	++_rank;
      else
	_rank = _rank+2*R+1;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_cont<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_cont<T,R> _tmp = *this;
      if((_rank+1)%_width<(_width-R))
	++_rank;
      else
	_rank = _rank+2*R+1;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,R>& operator=(iterator_cont<T,R> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      _width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_cont<T,R>& toTest) {return _rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_cont<T,R>& toTest) {return _rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_cont<T,R>& toComp) {return _rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_cont<T,R>& toComp) {return _rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_cont<T,R>& toComp) {return _rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_cont<T,R>& toComp) {return _rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //!  Iterator class
  /*!
    template of the contiguous iterator of SkelGIS. This is the second specialization of iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    R is specialized with the value 0
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_cont<T,0>
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_cont():_rank(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_cont(const unsigned int r):_rank(r){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_cont(const iterator_cont<T,0>& it):_rank(it._rank){}
    //------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_cont(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,0>& operator++(){++_rank;return *this;}
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,0>& operator++(int){iterator_cont<T,0> _tmp=*this;++_rank;return _tmp;}
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,0>& operator=(iterator_cont<T,0> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_cont<T,0>& toTest) {return _rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_cont<T,0>& toTest) {return _rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_cont<T,0>& toComp) {return _rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_cont<T,0>& toComp) {return _rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_cont<T,0>& toComp) {return _rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_cont<T,0>& toComp) {return _rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };
}

#endif




