/*! \file iterator_line.hpp
 *  \brief line iterator object definition. Two template specializations. This iterator gives the possibility to read elements of a DMAtrix line by line. Users do not have to construct an iterator object directly (these objects are constructed in the DMatrix) however operators and methods of the iterator are used by the user.
 */
#ifndef ITERATOR_LINE_H
#define ITERATOR_LINE_H

#include "iterator_cont.hpp"
#include "iterator.hpp"

namespace skelgis{

  //================================================================================
  //!  Iterator class
  /*!
    template of the line iterator of SkelGIS. This is the first specialization of line iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_line
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_line():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_line(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_line(const iterator_line<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_line(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      _rank = _rank + _width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_line<T,R> _tmp = *this;
      _rank = _rank + _width;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator -- pre-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,R>& operator--()
    //-------------------------------------------------------------------------------
    {
      _rank = _rank - _width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator -- post-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,R>& operator--(int)
    //-------------------------------------------------------------------------------
    {
      iterator_line<T,R> _tmp = *this;
      _rank = _rank - _width;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_line<T,R>& operator=(iterator_line<T,R> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      _width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_line<T,R>& toTest) {return _rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_line<T,R>& toTest) {return _rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_line<T,R>& toComp) {return _rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_line<T,R>& toComp) {return _rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_line<T,R>& toComp) {return _rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_line<T,R>& toComp) {return _rank>toComp._rank;}
    //-------------------------------------------------------------------------------
    //! begin method of the line iterator
    /*!
      This method returns an iterator on the first element of the current line of the DMatrix. This iterator is then used until the end() method of the line iterator is reached.
      \return an iterator on the first element of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> begin(){return iterator<T,R>(_rank,_width);}
    //-------------------------------------------------------------------------------
    //! begin_cont method of the line iterator
    /*!
      This method returns an iterator contiguous on the first element of the current line of the DMatrix. This iterator is then used until the end() method of the line iterator is reached.
      \return an iterator_cont on the first element of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,R> begin_cont(){return iterator_cont<T,R>(_rank,_width);}
    //-------------------------------------------------------------------------------
    //! end method of the line iterator
    /*!
      This method returns the end iterator on the current line of the DMatrix.
      \return the end iterator of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> end(){return iterator<T,R>(_rank+_width-2*R-1,_width);}
    //-------------------------------------------------------------------------------
    //! end_cont method of the line iterator
    /*!
      This method returns the end contiguous iterator on the current line of the DMatrix.
      \return the end_cont iterator of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,R> end_cont(){return iterator_cont<T,R>(_rank+_width-2*R-1,_width);}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //!  Iterator class
  /*!
    template of the line iterator of SkelGIS. This is the first specialization of line iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    R is specialized with the value 0
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_line<T,0>
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_line():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_line(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_line(const iterator_line<T,0>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_line(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,0>& operator++()
    //-------------------------------------------------------------------------------
    {
      _rank = _rank + _width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,0>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_line<T,0> _tmp = *this;
      _rank = _rank + _width;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator -- pre-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,0>& operator--()
    //-------------------------------------------------------------------------------
    {
      _rank = _rank - _width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator -- post-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_line<T,0>& operator--(int)
    //-------------------------------------------------------------------------------
    {
      iterator_line<T,0> _tmp = *this;
      _rank = _rank - _width;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_line<T,0>& operator=(iterator_line<T,0> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      _width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_line<T,0>& toTest) {return _rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_line<T,0>& toTest) {return _rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_line<T,0>& toComp) {return _rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_line<T,0>& toComp) {return _rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_line<T,0>& toComp) {return _rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_line<T,0>& toComp) {return _rank>toComp._rank;}
    //-------------------------------------------------------------------------------
    //! begin method of the line iterator
    /*!
      This method returns an iterator on the first element of the current line of the DMatrix. This iterator is then used until the end() method of the line iterator is reached.
      \return an iterator on the first element of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> begin(){return iterator<T,0>(_rank);}
    //-------------------------------------------------------------------------------
    //! begin_cont method of the line iterator
    /*!
      This method returns an iterator contiguous on the first element of the current line of the DMatrix. This iterator is then used until the end() method of the line iterator is reached.
      \return an iterator_cont on the first element of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,0> begin_cont(){return iterator_cont<T,0>(_rank);}
    //-------------------------------------------------------------------------------
    //! end method of the line iterator
    /*!
      This method returns the end iterator on the current line of the DMatrix.
      \return the end iterator of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> end(){return iterator<T,0>(_rank+_width-1);}
    //-------------------------------------------------------------------------------
    //! end_cont method of the line iterator
    /*!
      This method returns the end contiguous iterator on the current line of the DMatrix.
      \return the end_cont iterator of the current line on the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline iterator_cont<T,0> end_cont(){return iterator_cont<T,0>(_rank+_width-1);}
    //-------------------------------------------------------------------------------
  };

}

#endif
