/*! \file iterator_step.hpp
 *  \brief step iterator object definition. Two template specializations. This iterator gives the possibility to move by step in the DMatrix. _nb contiguous elements in the DMatrix will be read, and then _step elements will be jumped. Users do not have to construct an iterator object directly (these objects are constructed in the DMatrix) however operators and methods of the iterator are used by the user.
 */
#ifndef ITERATOR_STEP_H
#define ITERATOR_STEP_H

namespace skelgis{

  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_step
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */
    int _step; /*!< number of elements to jump after a read of _nb elements */
    int _nb; /*!< number of contiguous elements to read */
    int _cpt; /*!< counter to know how much elements has been read for know */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_step():_rank(0),_width(0),_cpt(1),_nb(1),_step(1){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
      \param s is the step value to jump
      \param n is the number of elements to read before the jump
    */
    //-------------------------------------------------------------------------------
    iterator_step(const unsigned int r,const unsigned int w,int s=1,int n=1):_rank(r),_width(w),_cpt(n),_nb(n),_step(s){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_step(const iterator_step<T,R>& it):_rank(it._rank),_width(it._width),_cpt(it._nb),_nb(it._nb),_step(it._step){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_step(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_step<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if(_cpt>1)
	{
	  if((_rank+1)%_width<(_width-R))
	    ++_rank;
	  else
	    _rank = _rank+2*R+1;

	  --_cpt;
	}
      else
	{
	  if((_rank+_step)%_width>=(_width-R) || (_rank+_step)%_width<_rank%_width)
	    _rank = _rank+_step+1+2*R+1;
	  else
	    _rank = _rank+_step+1;
	  _cpt = _nb;
	}
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_step<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_step<T,R> _tmp = *this;
      if(_cpt>1)
	{
	  if((_rank+1)%_width<(_width-R))
	    ++_rank;
	  else
	    _rank = _rank+2*R+1;

	  --_cpt;
	}
      else
	{
	  if((_rank+_step)%_width>=(_width-R) || (_rank+_step)%_width<_rank%_width)
	    _rank = _rank+_step+1+2*R+1;
	  else
	    _rank = _rank+_step+1;

	  _cpt = _nb;
	}
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_step<T,R>& operator=(iterator_step<T,R> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      _width = right._width;
      _cpt = right._nb;
      _nb = right._nb;
      _step = right._step;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_step<T,R> & toTest) {return this->_rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_step<T,R>& toTest) {return this->_rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_step<T,R>& toComp) {return this->_rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_step<T,R>& toComp) {return this->_rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_step<T,R>& toComp) {return this->_rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_step<T,R>& toComp) {return this->_rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //!  Iterator class
  /*!
    template of the step iterator of SkelGIS. This is the second specialization of iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    R is specialized with the value 0
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_step<T,0>
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    int _step; /*!< number of elements to jump after a read of _nb elements */
    int _nb; /*!< number of contiguous elements to read */
    int _cpt; /*!< counter to know how much elements has been read for know */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_step():_rank(0),_cpt(1),_nb(1),_step(1){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
      \param s is the step value to jump
      \param n is the number of elements to read before the jump
    */
    //-------------------------------------------------------------------------------
    iterator_step(const unsigned int r,int s=1,int n=1):_rank(r),_cpt(n),_nb(n),_step(s){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_step(const iterator_step<T,0>& it):_rank(it._rank),_cpt(it._nb),_nb(it._nb),_step(it._step){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_step(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_step<T,0>& operator++()
    //-------------------------------------------------------------------------------
    {
      if(_cpt>1)
	{
	  ++_rank;
	  --_cpt;
	}
      else
	{
	  _rank = _rank+_step+1;
	  _cpt = _nb;
	}
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_step<T,0>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_step<T,0> _tmp = *this;
      if(_cpt>1)
	{
	  ++_rank;
	  --_cpt;
	}
      else
	{
	  _rank = _rank+_step+1;
	  _cpt = _nb;
	}
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_step<T,0>& operator=(iterator_step<T,0> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      _cpt = right._nb;
      _nb = right._nb;
      _step = right._step;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_step<T,0> & toTest) {return this->_rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_step<T,0>& toTest) {return this->_rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_step<T,0>& toComp) {return this->_rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_step<T,0>& toComp) {return this->_rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_step<T,0>& toComp) {return this->_rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_step<T,0>& toComp) {return this->_rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

}

#endif
