/*! \file apply_binary.hpp
 *  \brief Definitions of the skeleton object ApplyBinary. This skeleton takes a user function and two DMatrix as inputs and return a new DMatrix as output.
 */
#ifndef APPLY_BINARY_H
#define APPLY_BINARY_H

namespace skelgis{

  //================================================================================
  //Types of user functions
  //================================================================================
  //! _ApplyBinary_Func structure
  /*!
    These objects are not directly used by the user, but the macros used by the user, to define its functions, uses these. Four specializations of this structure are available.
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation for the two input DMatrix. It has to be the same for the two inputs.
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template <typename T,int R, typename T2, typename T3,int R3,bool line=true> struct _ApplyBinary_Func
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix <T,R>&, DMatrix<T2,R>&, DMatrix<T3,R3>&) const =0;
  };
  //================================================================================
  //! _ApplyBinary_Func structure
  /*!
    \tparam T is the type of data to store in the first input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
    The default value of line is used (true).
  */
  //-------------------------------------------------------------------------------
  template <typename T, typename T2, typename T3,int R3> struct _ApplyBinary_Func<T,0,T2,T3,R3>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix <T,0>&, DMatrix<T2,0>&,DMatrix<T3,R3>&) const =0;
  };
  //================================================================================
  //! _ApplyBinary_Func structure
  /*!
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation for the two input DMatrix. It has to be the same for the two inputs.
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template <typename T,int R, typename T2, typename T3,int R3> struct _ApplyBinary_Func<T,R,T2,T3,R3,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix <T,R,false>&, DMatrix<T2,R,false>&, DMatrix<T3,R3,false>&) const =0;
  };
  //================================================================================
  //! _ApplyBinary_Func structure
  /*!
    \tparam T is the type of data to store in the first input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template <typename T, typename T2, typename T3,int R3> struct _ApplyBinary_Func<T,0,T2,T3,R3,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix <T,0,false>&, DMatrix<T2,0,false>&,DMatrix<T3,R3,false>&) const =0;
  };
  //-------------------------------------------------------------------------------


  //================================================================================
  //Specializations of skeleton ApplyBinary
  //================================================================================
  //! ApplyBinary structure
  /*!
    These objects are directly called by the user to apply a function on two inputs DMatrix.
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation for the two input DMatrix. It has to be the same for the two inputs.
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
    \tparam local is a particular choice if it is explicitely indicate to not make any MPI exchanges. The default behavior is to make normaly MPI exchanges.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,class T3,int R3,bool line=true,bool local=false> struct ApplyBinary{};
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    First specialization with line and local default values.
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation for the two input DMatrix. It has to be the same for the two inputs.
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,class T3,int R3> struct ApplyBinary<T,R,T2,T3,R3>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3> apply(const _ApplyBinary_Func<T,R,T2,T3,R3>& func, DMatrix<T,R>& m, DMatrix<T2,R>& m2)
    {
      DMatrix<T3,R3> m3(m.getGlobalHeader(),0);
      m.getBorders();
      m2.getBorders();
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,R,T2,T3,R3>& func, DMatrix<T,R>& m, DMatrix<T2,R>& m2, DMatrix<T3,R3>& m3)
    {
      m.getBorders();
      m2.getBorders();
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Second specialization with line and local default values and R=0
    \tparam T is the type of data to store in the first input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,class T3,int R3> struct ApplyBinary<T,0,T2,T3,R3>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3> apply(const _ApplyBinary_Func<T,0,T2,T3,R3>& func, DMatrix<T,0>& m, DMatrix<T2,0>& m2)
    {
      DMatrix<T3,R3> m3(m.getGlobalHeader(),0);
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,0,T2,T3,R3>& func, DMatrix<T,0>& m, DMatrix<T2,0>& m2, DMatrix<T3,R3>& m3)
    {
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Third specialization with line=true and local=true
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation for the two input DMatrix. It has to be the same for the two inputs.
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,class T3,int R3> struct ApplyBinary<T,R,T2,T3,R3,true,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3> apply(const _ApplyBinary_Func<T,R,T2,T3,R3>& func, DMatrix<T,R>& m, DMatrix<T2,R>& m2)
    {
      DMatrix<T3,R3> m3(m.getGlobalHeader(),0);
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,R,T2,T3,R3>& func, DMatrix<T,R>& m, DMatrix<T2,R>& m2, DMatrix<T3,R3>& m3)
    {
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Fourth specialization with line=true and local=true and R=0
    \tparam T is the type of data to store in the first input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,class T3,int R3> struct ApplyBinary<T,0,T2,T3,R3,true,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3> apply(const _ApplyBinary_Func<T,0,T2,T3,R3>& func, DMatrix<T,0>& m, DMatrix<T2,0>& m2)
    {
      DMatrix<T3,R3> m3(m.getGlobalHeader(),0);
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,0,T2,T3,R3>& func, DMatrix<T,0>& m, DMatrix<T2,0>& m2, DMatrix<T3,R3>& m3)
    {
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Fith specialization with line=false and local=false default value
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation, in other words, the size of the physical border needed
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,class T3,int R3> struct ApplyBinary<T,R,T2,T3,R3,false>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3,false> apply(const _ApplyBinary_Func<T,R,T2,T3,R3,false>& func, DMatrix<T,R,false>& m, DMatrix<T2,R,false>& m2)
    {
      DMatrix<T3,R3,false> m3(m.getGlobalHeader(),0);
      m.getBorders();
      m2.getBorders();
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,R,T2,T3,R3,false>& func, DMatrix<T,R,false>& m, DMatrix<T2,R,false>& m2, DMatrix<T3,R3,false>& m3)
    {
      m.getBorders();
      m2.getBorders();
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Sixth specialization with line=false and local=false default value and R=0
    \tparam T is the type of data to store in the first input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,class T3,int R3> struct ApplyBinary<T,0,T2,T3,R3,false>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3,false> apply(const _ApplyBinary_Func<T,0,T2,T3,R3,false>& func, DMatrix<T,0,false>& m, DMatrix<T2,0,false>& m2)
    {
      DMatrix<T3,R3,false> m3(m.getGlobalHeader(),0);
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,0,T2,T3,R3,false>& func, DMatrix<T,0,false>& m, DMatrix<T2,0,false>& m2, DMatrix<T3,R3,false>& m3)
    {
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Seventh specialization with line=false and local=true
    \tparam T is the type of data to store in the first input DMatrix
    \tparam R is the overlap distance needed by the calculation, in other words, the size of the physical border needed
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,class T3,int R3> struct ApplyBinary<T,R,T2,T3,R3,false,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3,false> apply(const _ApplyBinary_Func<T,R,T2,T3,R3,false>& func, DMatrix<T,R,false>& m, DMatrix<T2,R,false>& m2)
    {
      DMatrix<T3,R3,false> m3(m.getGlobalHeader(),0);
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,R,T2,T3,R3,false>& func, DMatrix<T,R,false>& m, DMatrix<T2,R,false>& m2, DMatrix<T3,R3,false>& m3)
    {
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyBinary structure
  /*!
    Eighth specialization with line=false and local=true and R=0
    \tparam T is the type of data to store in the first input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the second inputDMatrix
    \tparam T3 is the type of data to store in the output DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,class T3,int R3> struct ApplyBinary<T,0,T2,T3,R3,false,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T3,R3,false> apply(const _ApplyBinary_Func<T,0,T2,T3,R3,false>& func, DMatrix<T,0,false>& m, DMatrix<T2,0,false>& m2)
    {
      DMatrix<T3,R3,false> m3(m.getGlobalHeader(),0);
      func(m,m2,m3);
      return m3;
      }*/

    inline static void apply(const _ApplyBinary_Func<T,0,T2,T3,R3,false>& func, DMatrix<T,0,false>& m, DMatrix<T2,0,false>& m2, DMatrix<T3,R3,false>& m3)
    {
      func(m,m2,m3);
    }
  };
  //-------------------------------------------------------------------------------

}

#endif
