/*! \file apply_list.hpp
 *  \brief Definitions of the skeleton object ApplyList. This skeleton takes a list of DMatrix and a user function as inputs and returns a list of DMatrix as output.
 */
#ifndef APPLYLIST_H
#define APPLYLIST_H

#include "../../../util/dmatrix_vector.hpp"

namespace skelgis{

  //================================================================================
  //Types of user functions
  //================================================================================
  //! _ApplyList_Func structure
  /*!
    These objects are not directly used by the user, but the macros used by the user, to define its functions, uses these. Four specializations of this structure are available.
    \tparam T is the type of data to store in the input list of DMatrix
    \tparam R is the overlap distance needed by the calculation for input DMatrix. 
    \tparam T2 is the type of data to store in the output list of DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix. 
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template <typename T,int R, typename T2, int R2,bool line=true> struct _ApplyList_Func
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix_vector<T,R>&,DMatrix_vector<T2,R2>&) const =0;
  };
  //================================================================================
  //! _ApplyList_Func structure
  /*!
    \tparam T is the type of data to store in the input list of DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output list of DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix. 
    The default value of line is used (true).
  */
  //-------------------------------------------------------------------------------
  template <typename T, typename T2, int R2> struct _ApplyList_Func<T,0,T2,R2>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix_vector<T,0>&, DMatrix_vector<T2,R2>&) const =0;
  };
  //================================================================================
  //! _ApplyList_Func structure
  /*!
    \tparam T is the type of data to store in the input list of DMatrix
    \tparam R is the overlap distance needed by the calculation for input DMatrix. 
    \tparam T2 is the type of data to store in the output list of DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix. 
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block)
  */
  //-------------------------------------------------------------------------------
  template <typename T,int R, typename T2,int R2> struct _ApplyList_Func<T,R,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix_vector<T,R,false>&,DMatrix_vector<T2,R2,false>&) const =0;
  };
  //================================================================================
  //! _ApplyList_Func structure
  /*!
    \tparam T is the type of data to store in the input list of DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output list of DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template <typename T, typename T2,int R2> struct _ApplyList_Func<T,0,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix_vector<T,0,false>&, DMatrix_vector<T2,R2,false>&) const =0;
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //Specializations of skeleton ApplyList
  //================================================================================
  //! ApplyList structure
  /*!
    These objects are directly called by the user to apply a function on a list of Dmatrix.
    \tparam T is the type of data to store in input list of Dmatrix
    \tparam R is the overlap distance needed by the calculation
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
    \tparam local is a particular choice if it is explicitely indicate to not make any MPI exchanges. The default behavior is to make normaly MPI exchanges.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2, int R2,bool line=true,bool local=false> struct ApplyList{};
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    First specialization with line and local default values.
    \tparam T is the type of data to store in input list of Dmatrix
    \tparam R is the overlap distance needed by the calculation
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyList<T,R,T2,R2>
  //-------------------------------------------------------------------------------
  {
    static void apply(const _ApplyList_Func<T,R,T2,R2>& func,DMatrix_vector<T,R>& inputs, DMatrix_vector<T2,R2>& outputs)
    {
      for(int i=0;i<inputs.size();i++)
	{
	  DMatrix_impl<T,R> * m = inputs[i];
	  m->getBorders();
	}
      func(inputs,outputs);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Second specialization with line and local default values and R=0
    \tparam T is the type of data to store in input list of Dmatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyList<T,0,T2,R2>
  //-------------------------------------------------------------------------------
  {
    inline static void apply(const _ApplyList_Func<T,0,T2,R2>& func,DMatrix_vector<T,0>& inputs, DMatrix_vector<T2,R2>& outputs){func(inputs,outputs);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Third specialization with with line=true and local=true
    \tparam T is the type of data to store in input list of Dmatrix
    \tparam R is the overlap distance needed by the calculation
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyList<T,R,T2,R2,true,true>
  //-------------------------------------------------------------------------------
  {
    inline static void apply(const _ApplyList_Func<T,R,T2,R2>& func, DMatrix_vector<T,R>& inputs, DMatrix_vector<T2,R2>& outputs){func(inputs,outputs);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Fourth specialization with line=true and local=true and R=0
    \tparam T is the type of data to store in input list of Dmatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyList<T,0,T2,R2,true,true>
  //-------------------------------------------------------------------------------
  {
    inline static void apply(const _ApplyList_Func<T,0,T2,R2>& func, DMatrix_vector<T,0>& inputs, DMatrix_vector<T2,R2>& outputs){func(inputs,outputs);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Fith specialization with line=false and local=false default value
    \tparam T is the type of data to store in input list of Dmatrix
    \tparam R is the overlap distance needed by the calculation
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyList<T,R,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    static void apply(const _ApplyList_Func<T,R,T2,R2,false>& func,DMatrix_vector<T,R,false>& inputs, DMatrix_vector<T2,R2,false>& outputs)
    {
      for(int i=0;i<inputs.size();i++)
	{
	  DMatrix_impl<T,R,false> * m = inputs[i];
	  m->getBorders();
	}
      func(inputs,outputs);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Sixth specialization with line=false and local=false default value and R=0
    \tparam T is the type of data to store in input list of Dmatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyList<T,0,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    inline static void apply(const _ApplyList_Func<T,0,T2,R2,false>& func,DMatrix_vector<T,0,false>& inputs, DMatrix_vector<T2,R2,false>& outputs){func(inputs,outputs);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Seventh specialization with line=false and local=true
    \tparam T is the type of data to store in input list of Dmatrix
    \tparam R is the overlap distance needed by the calculation
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyList<T,R,T2,R2,false,true>
  //-------------------------------------------------------------------------------
  {
    inline static void apply(const _ApplyList_Func<T,R,T2,R2,false>& func, DMatrix_vector<T,R,false>& inputs, DMatrix_vector<T2,R2,false>& outputs){func(inputs,outputs);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyList structure
  /*!
    Eighth specialization with line=false and local=true and R=0
    \tparam T is the type of data to store in input list of Dmatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyList<T,0,T2,R2,false,true>
  //-------------------------------------------------------------------------------
  {
    inline static void apply(const _ApplyList_Func<T,0,T2,R2,false>& func, DMatrix_vector<T,0,false>& inputs, DMatrix_vector<T2,R2,false>& outputs){func(inputs,outputs);}
  };
  //-------------------------------------------------------------------------------
  
}

#endif
