/*! \file apply_unary.hpp
 *  \brief Definitions of the skeleton object ApplyUnary. This skeleton takes a user function and a DMatrix as inputs and return a new DMatrix as output.
 */
#ifndef APPLY_UNARY_H
#define APPLY_UNARY_H

namespace skelgis{

  //================================================================================
  //Types of user functions
  //================================================================================
  //! _ApplyUnary_Func structure
  /*!
    These objects are not directly used by the user, but the macros used by the user, to define its functions, use these. Four specializations of this structure are available.
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template <typename T,int R, typename T2,int R2,bool line=true> struct _ApplyUnary_Func
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix <T,R>&, DMatrix<T2,R2>&) const =0;
  };
  //================================================================================
  //! _ApplyUnary_Func structure
  /*!
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
    The default value of line is used (true)
  */
  //-------------------------------------------------------------------------------
  template <typename T, typename T2,int R2> struct _ApplyUnary_Func<T,0,T2,R2>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix<T,0>&, DMatrix<T2,R2>&) const =0;
  };
  //================================================================================
  //! _ApplyUnary_Func structure
  /*!
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block)
  */
  //-------------------------------------------------------------------------------
  template <typename T, int R, typename T2,int R2> struct _ApplyUnary_Func<T,R,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix<T,R,false>&, DMatrix<T2,R2,false>&) const =0;
  };
  //================================================================================
  //! _ApplyUnary_Func structure
  /*!
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block)
  */
  //-------------------------------------------------------------------------------
  template <typename T, typename T2,int R2> struct _ApplyUnary_Func<T,0,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix<T,0,false>&, DMatrix<T2,R2,false>&) const =0;
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //Specializations of skeleton ApplyUnary
  //================================================================================
  //! ApplyUnary structure
  /*!
    These objects are directly called by the user to apply a function on a DMatrix.
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
    \tparam local is a particular choice if it is explicitely indicate to not make any MPI exchanges. The default behavior is to make normaly MPI exchanges.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2,bool line=true,bool local=false> struct ApplyUnary{};
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    First specialization with line and local default values.
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyUnary<T,R,T2,R2>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2> apply(const _ApplyUnary_Func<T,R,T2,R2>& func, DMatrix<T,R>& m)
    {
      DMatrix<T2,R2> m2(m.getGlobalHeader(),0);
      m.getBorders();
      func(m,m2);
      return m2;
      }*/

    static void apply(const _ApplyUnary_Func<T,R,T2,R2>& func, DMatrix<T,R>& m, DMatrix<T2,R2>& m2)
    {
      m.getBorders();
      func(m,m2);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Second specialization with line and local default values and R=0
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyUnary<T,0,T2,R2>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2> apply(const _ApplyUnary_Func<T,0,T2,R2>& func, DMatrix<T,0>& m)
    {
      DMatrix<T2,R2> m2(m.getGlobalHeader(),0);
      func(m,m2);
      return m2;
      }*/

    inline static void apply(const _ApplyUnary_Func<T,0,T2,R2>& func, DMatrix<T,0>& m, DMatrix<T2,R2>& m2){func(m,m2);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Third specialization with line=true and local=true
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyUnary<T,R,T2,R2,true,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2> apply(const _ApplyUnary_Func<T,R,T2,R2>& func, DMatrix<T,R>& m)
    {
      DMatrix<T2,R2> m2(m.getGlobalHeader(),0);
      func(m,m2);
      return m2;
      }*/

    inline static void apply(const _ApplyUnary_Func<T,R,T2,R2>& func, DMatrix<T,R>& m, DMatrix<T2,R2>& m2){func(m,m2);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Fourth specialization with line=true and local=true and R=0
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyUnary<T,0,T2,R2,true,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2> apply(const _ApplyUnary_Func<T,0,T2,R2>& func, DMatrix<T,0>& m)
    {
      DMatrix<T2,R2> m2(m.getGlobalHeader(),0);
      func(m,m2);
      return m2;
      }*/

    inline static void apply(const _ApplyUnary_Func<T,0,T2,R2>& func, DMatrix<T,0>& m, DMatrix<T2,R2>& m2){func(m,m2);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Fith specialization with line=false and local=false default value
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyUnary<T,R,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2,false> apply(const _ApplyUnary_Func<T,R,T2,R2,false>& func, DMatrix<T,R,false>& m)
    {
      DMatrix<T2,R2,false> m2(m.getGlobalHeader(),0);
      m.getBorders();
      func(m,m2);
      return m2;
      }*/

    static void apply(const _ApplyUnary_Func<T,R,T2,R2,false>& func, DMatrix<T,R,false>& m, DMatrix<T2,R2,false>& m2)
    {
      m.getBorders();
      func(m,m2);
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Sixth specialization with line=false and local=false default value and R=0
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyUnary<T,0,T2,R2,false>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2,false> apply(const _ApplyUnary_Func<T,0,T2,R2,false>& func, DMatrix<T,0,false>& m)
    {
      DMatrix<T2,R2,false> m2(m.getGlobalHeader(),0);
      func(m,m2);
      return m2;
      }*/

    inline static void apply(const _ApplyUnary_Func<T,0,T2,R2,false>& func, DMatrix<T,0,false>& m, DMatrix<T2,R2,false>& m2){func(m,m2);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Seventh specialization with line=false and local=true
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,class T2,int R2> struct ApplyUnary<T,R,T2,R2,false,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2,false> apply(const _ApplyUnary_Func<T,R,T2,R2,false>& func, DMatrix<T,R,false>& m)
    {
      DMatrix<T2,R2,false> m2(m.getGlobalHeader(),0);
      func(m,m2);
      return m2;
      }*/

    inline static void apply(const _ApplyUnary_Func<T,R,T2,R2,false>& func, DMatrix<T,R,false>& m, DMatrix<T2,R2,false>& m2){func(m,m2);}
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyUnary structure
  /*!
    Eighth specialization with line=false and local=true and R=0
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam T2 is the type of data to store in the output DMatrix
    \tparam R2 is the overlap distance needed by the calculation for the output DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,class T2,int R2> struct ApplyUnary<T,0,T2,R2,false,true>
  //-------------------------------------------------------------------------------
  {
    //need an operator=(DMatrixw<...> m) in dmatrix.hpp
    /*inline static DMatrix<T2,R2,false> apply(const _ApplyUnary_Func<T,0,T2,R2,false>& func, DMatrix<T,0,false>& m)
    {
      DMatrix<T2,R2,false> m2(m.getGlobalHeader(),0);
      func(m,m2);
      return m2;
      }*/

    inline static void apply(const _ApplyUnary_Func<T,0,T2,R2,false>& func, DMatrix<T,0,false>& m, DMatrix<T2,R2,false>& m2){func(m,m2);}
    };
  //-------------------------------------------------------------------------------

}

#endif
