/*! \file communications.hpp
 *  \brief Definitions of the communication class of SkelGIS. This class is used to make MPI communications.
 */
#ifndef COMMUNICATIONS_H
#define COMMUNICATIONS_H

#include "mpi.h"
#include "mpi_.hpp"

namespace skelgis{

  //-------------------------------------------------------------------------------
  //!  template communication Class 
  /*!
    This template class defines the MPI initializations, the MPI communications
    \tpara T is the type of data to exchange
  */
  //-------------------------------------------------------------------------------
  template<class T> class Communications
  //-------------------------------------------------------------------------------
  {
  public:
    //-------------------------------------------------------------------------------
    //! used by get borders to make exchanges
    /*!
      \param toGet is the data to get from mpi process and omp thread
      \param toSend is the data to send to mpi process and omp thread
      \param size is the size of data to send and recv
      \param rank is the mpi rank to exchange with
    */
    //-------------------------------------------------------------------------------
    static void Exchanges(T * toSend, T* toGet,unsigned int size,int rank)
    //-------------------------------------------------------------------------------
    {
      MPI_Status statu;
      MPI_Request request;
      
#ifdef DEBUG
      std::stringstream st;
      st<<"Values to give to rank "<<rank<<"\n";
      for(int j=0;j<size;j++)
	{
	  st<<toSend[j]<<" ";
	}
      st<<"\n";
      Mpi_::print(st.str());
#endif

      //work for every basic types not for complex types
      char * get = new char[size*sizeof(T)];
      MPI_Sendrecv(reinterpret_cast<char *>(toSend), size*(sizeof(T)), MPI_BYTE,rank,0,get,size*(sizeof(T)),MPI_BYTE,rank,0,MPI_COMM_WORLD,&statu);
      for(int i=0;i<size;i++)
	  memcpy(&toGet[i], get+i*sizeof(T), sizeof(T));
      delete [] get;

#ifdef DEBUG
      std::stringstream st2;
      st2<<"values given by rank "<<rank<<"\n";
      for(int j=0;j<size;j++)
	{
	  st2<<toGet[j]<<" ";
	}
      st2<<"\n";
      Mpi_::print(st2.str());
#endif
    }
    //-------------------------------------------------------------------------------

    
    //-------------------------------------------------------------------------------
    //for DDAG construction of an object needed
    //an instance of the object is constructed for each DPMap (nodes or edges)
    //this object manage non blocking MPI exchanges and received values as a MPI_BYTE type
    //-------------------------------------------------------------------------------

    /*!< requests_s is the list of requests for send communications (one for each processor) */
    MPI_Request * requests_s;
    /*!< requests_s is the list of requests for receive communications (one for each processor) */
    MPI_Request * requests_r;
    /*!< requests_s is the list of status for send communications (one for each processor) */
    MPI_Status * status_s;
    /*!< requests_s is the list of status for receive communications (one for each processor) */
    MPI_Status * status_r;
    char ** get;

    //-------------------------------------------------------------------------------
    //! constructor
    /*!
    */
    //-------------------------------------------------------------------------------
    Communications()
    //-------------------------------------------------------------------------------
    {
      requests_s = (MPI_Request *)calloc(Mpi_::mpi_nb,sizeof(MPI_Request));
      requests_r = (MPI_Request *)calloc(Mpi_::mpi_nb,sizeof(MPI_Request));
      status_s = (MPI_Status *)calloc(Mpi_::mpi_nb,sizeof(MPI_Status));
      status_r = (MPI_Status *)calloc(Mpi_::mpi_nb,sizeof(MPI_Status));
      get = (char **)calloc(Mpi_::mpi_nb,sizeof(char*));
    }
    //-------------------------------------------------------------------------------
    //! destructor
    /*!
    */
    //-------------------------------------------------------------------------------
    ~Communications()
    //-------------------------------------------------------------------------------
    {
      free(requests_s);
      free(requests_r);
      free(status_s);
      free(status_r);
      free(get);
    }
    //-------------------------------------------------------------------------------
    //! A non blocking MPI send is initialized
    /*!
      \param buf is the sending buffer
      \param count is the number of elements in the buffer
      \param dest is the destination processor
      \return the return value of the MPI_Isend function
    */
    //-------------------------------------------------------------------------------
    inline int ISend(T *buf, int count, int dest)
    //-------------------------------------------------------------------------------
    {
      return  MPI_Isend(reinterpret_cast<char *>(buf), count*(sizeof(T)), MPI_BYTE, dest, 0, MPI_COMM_WORLD, &(requests_s[dest]));
    }
    //-------------------------------------------------------------------------------
    //! A non blocking MPI receive is initialized
    /*!
      \param count is the number of elements to receive
      \param source is the source processor
      \return the return value of the MPI_Irecv function
    */
    //-------------------------------------------------------------------------------
    int IRecv(int count, int source)
    //-------------------------------------------------------------------------------
    {
      get[source] = new char[count*sizeof(T)];
      return MPI_Irecv(get[source], count*(sizeof(T)), MPI_BYTE, source, 0, MPI_COMM_WORLD, &(requests_r[source]));
    }
    //-------------------------------------------------------------------------------
    //! A MPI wait call to wait for the send exchange with a precise processor
    /*!
      \param dest is the destination processor the send is waiting
      \return the return value of the MPI_Wait function
    */
    //-------------------------------------------------------------------------------
    inline int Wait_s(int dest)
    //-------------------------------------------------------------------------------
    {
      return MPI_Wait(&(requests_s[dest]), &(status_s[dest]));
    }
    //-------------------------------------------------------------------------------
    //! A MPI wait call to wait for the send exchange with a precise processor
    /*!
      \param dest is the destination processor the send is waiting
      \return the return value of the MPI_Wait function
    */
    //-------------------------------------------------------------------------------
    int Wait_r(T * toRecv, int src,int count)
    //-------------------------------------------------------------------------------
    {
      int ret = MPI_Wait(&(requests_r[src]), &(status_r[src]));
      for(int i=0;i<count;i++)
	memcpy(&toRecv[i], get[src]+i*sizeof(T), sizeof(T));

      delete [] get[src];

      return ret;
    }
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------


  //-------------------------------------------------------------------------------
  //!  template communicationInOut Class 
  /*!
    This template class defines the MPI initializations, the MPI communications
    \tpara T is the type of data to exchange
  */
  //-------------------------------------------------------------------------------
  template<class T> class CommunicationsInOut
  //-------------------------------------------------------------------------------
  {
  public:
    //-------------------------------------------------------------------------------
    //for DDAG construction of an object needed
    //an instance of the object is constructed for each DPMap (nodes or edges) advanced on pointers
    //this object manage non blocking MPI exchanges and received values as a MPI_BYTE type
    //-------------------------------------------------------------------------------

    /*!< requests_s is the list of requests for send communications (one for each processor) */
    MPI_Request * requests_s_out;
    MPI_Request * requests_s_in;
    /*!< requests_s is the list of requests for receive communications (one for each processor) */
    MPI_Request * requests_r_out;
    MPI_Request * requests_r_in;
    /*!< requests_s is the list of status for send communications (one for each processor) */
    MPI_Status * status_s_out;
    MPI_Status * status_s_in;
    /*!< requests_s is the list of status for receive communications (one for each processor) */
    MPI_Status * status_r_out;
    MPI_Status * status_r_in;
    char ** get_out;
    char ** get_in;

    //-------------------------------------------------------------------------------
    //! constructor
    /*!
    */
    //-------------------------------------------------------------------------------
    CommunicationsInOut()
    //-------------------------------------------------------------------------------
    {
      requests_s_out = (MPI_Request *)calloc(Mpi_::mpi_nb,sizeof(MPI_Request));
      requests_s_in = (MPI_Request *)calloc(Mpi_::mpi_nb,sizeof(MPI_Request));
      requests_r_out = (MPI_Request *)calloc(Mpi_::mpi_nb,sizeof(MPI_Request));
      requests_r_in = (MPI_Request *)calloc(Mpi_::mpi_nb,sizeof(MPI_Request));
      status_s_out = (MPI_Status *)calloc(Mpi_::mpi_nb,sizeof(MPI_Status));
      status_s_in = (MPI_Status *)calloc(Mpi_::mpi_nb,sizeof(MPI_Status));
      status_r_out = (MPI_Status *)calloc(Mpi_::mpi_nb,sizeof(MPI_Status));
      status_r_in = (MPI_Status *)calloc(Mpi_::mpi_nb,sizeof(MPI_Status));
      get_out = (char **)calloc(Mpi_::mpi_nb,sizeof(char*));
      get_in = (char **)calloc(Mpi_::mpi_nb,sizeof(char*));
    }
    //-------------------------------------------------------------------------------
    //! destructor
    /*!
    */
    //-------------------------------------------------------------------------------
    ~CommunicationsInOut()
    //-------------------------------------------------------------------------------
    {
      free(requests_s_out);
      free(requests_s_in);
      free(requests_r_out);
      free(requests_r_in);
      free(status_s_out);
      free(status_s_in);
      free(status_r_out);
      free(status_r_in);
      free(get_out);
      free(get_in);
    }
    //-------------------------------------------------------------------------------
    //! A non blocking MPI send is initialized
    /*!
      \param buf is the sending buffer out of another processor
      \param count is the number of elements in the buffer
      \param dest is the destination processor
      \return the return value of the MPI_Isend function
    */
    //-------------------------------------------------------------------------------
    inline int ISendOut(T *bufOut, int count, int dest)
    //-------------------------------------------------------------------------------
    {
      return  MPI_Isend(reinterpret_cast<char *>(bufOut), count*(sizeof(T)), MPI_BYTE, dest, 0, MPI_COMM_WORLD, &(requests_s_out[dest]));
    }
    //-------------------------------------------------------------------------------
    //! A non blocking MPI send is initialized
    /*!
      \param buf is the sending buffer in another processor
      \param count is the number of elements in the buffer
      \param dest is the destination processor
      \return the return value of the MPI_Isend function
    */
    //-------------------------------------------------------------------------------
    inline int ISendIn(T *bufIn, int count, int dest)
    //-------------------------------------------------------------------------------
    {
      return  MPI_Isend(reinterpret_cast<char *>(bufIn), count*(sizeof(T)), MPI_BYTE, dest, 0, MPI_COMM_WORLD, &(requests_s_in[dest]));
    }
    //-------------------------------------------------------------------------------
    //! A non blocking MPI receive is initialized
    /*!
      \param count is the number of elements to receive from an out edge
      \param source is the source processor
      \return the return value of the MPI_Irecv function
    */
    //-------------------------------------------------------------------------------
    int IRecvOut(int count, int source)
    //-------------------------------------------------------------------------------
    {
      get_out[source] = new char[count*sizeof(T)];
      return MPI_Irecv(get_out[source], count*(sizeof(T)), MPI_BYTE, source, 0, MPI_COMM_WORLD, &(requests_r_out[source]));
    }
    //-------------------------------------------------------------------------------
    //! A non blocking MPI receive is initialized
    /*!
      \param count is the number of elements to receive from an in edge
      \param source is the source processor
      \return the return value of the MPI_Irecv function
    */
    //-------------------------------------------------------------------------------
    int IRecvIn(int count, int source)
    //-------------------------------------------------------------------------------
    {
      get_in[source] = new char[count*sizeof(T)];
      return MPI_Irecv(get_in[source], count*(sizeof(T)), MPI_BYTE, source, 0, MPI_COMM_WORLD, &(requests_r_in[source]));
    }
    //-------------------------------------------------------------------------------
    //! A MPI wait call to wait for the send exchange with a precise processor
    /*!
      \param dest is the destination processor the send is waiting
      \return the return value of the MPI_Wait function
    */
    //-------------------------------------------------------------------------------
    inline int Wait_sOut(int dest)
    //-------------------------------------------------------------------------------
    {
      return MPI_Wait(&(requests_s_out[dest]), &(status_s_out[dest]));
    }
    //-------------------------------------------------------------------------------
    //! A MPI wait call to wait for the send exchange with a precise processor
    /*!
      \param dest is the destination processor the send is waiting
      \return the return value of the MPI_Wait function
    */
    //-------------------------------------------------------------------------------
    inline int Wait_sIn(int dest)
    //-------------------------------------------------------------------------------
    {
      return MPI_Wait(&(requests_s_in[dest]), &(status_s_in[dest]));
    }
    //-------------------------------------------------------------------------------
    //! A MPI wait call to wait for the send exchange with a precise processor
    /*!
      \param dest is the destination processor the send is waiting
      \return the return value of the MPI_Wait function
    */
    //-------------------------------------------------------------------------------
    int Wait_rOut(T * toRecvOut, int src,int count)
    //-------------------------------------------------------------------------------
    {
      int ret = MPI_Wait(&(requests_r_out[src]), &(status_r_out[src]));
      for(int i=0;i<count;i++)
	memcpy(&toRecvOut[i], get_out[src]+i*sizeof(T), sizeof(T));

      delete [] get_out[src];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //! A MPI wait call to wait for the send exchange with a precise processor
    /*!
      \param dest is the destination processor the send is waiting
      \return the return value of the MPI_Wait function
    */
    //-------------------------------------------------------------------------------
    int Wait_rIn(T * toRecvIn, int src,int count)
    //-------------------------------------------------------------------------------
    {
      int ret = MPI_Wait(&(requests_r_in[src]), &(status_r_in[src]));
      for(int i=0;i<count;i++)
	memcpy(&toRecvIn[i], get_in[src]+i*sizeof(T), sizeof(T));

      delete [] get_in[src];

      return ret;
    }
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------
}

#endif

