/*! \file macros.hpp
 *  \brief macros of SKelGIS library
 */
#ifndef MACROS_H
#define MACROS_H

#include "mpi.h"
#include "dmatrix_vector.hpp"

namespace skelgis{

//------------MPI FUNCS
//! MPI_RANK_NB macro
/*!
  Used to initialize the number of mpi processes and the rank of the current process
*/
#define MPI_RANK_NB MPI_Comm_size(MPI_COMM_WORLD,&(Mpi_::mpi_nb));	\
MPI_Comm_rank(MPI_COMM_WORLD,&(Mpi_::mpi_rank));
//! MPI_END macro
/*!
  Used to end MPI
*/
#define MPI_END MPI_Finalize();
//------------
//------------SKELGIS INIT
//! INITSKELGIS
/*!
  Used to initialize SkelGIS library use. Mpi initializations.
*/
#define INITSKELGIS MPI_Init(&argc,&argv); \
MPI_RANK_NB
//! ENDSKELGIS
/*!
  Used to end SkelGIS library use.
*/
//-----------SKELGIS END
#define ENDSKELGIS MPI_END
//---------- 

//----------PARAMETERS CLASS
#define BEGINParam(name) struct name{ \
private : \
 name(){}  \
 ~name(){} \
 static name * singleton; \
public: \
 static name * getInstance() \
 { \
 if(NULL==singleton) \
   singleton = new name(); \
 return singleton;	   \
 }

#define ENDParam(name) }; \
  name *name::singleton = NULL;
//----------
//----------GET PARAMETER
#define PARAM(class_name,param_name) class_name::getInstance()->param_name 
//----------

//==========================================
//============DMATRIX SKELETONS=============
//==========================================
//-----------SKELETON APPLYONE FUNCTION
//! BEGINApplyUnary
/*!
  Used to code the user function for the skeleton ApplyUnary
*/
#define BEGINApplyUnary(name, input_matrix, input_type, r, output_matrix, output_type, r2) struct _ApplyUnary_##name : public _ApplyUnary_Func<input_type, r, output_type,r2> \
{ \
  void operator()(DMatrix<input_type,r>& input_matrix, DMatrix<output_type,r2>& output_matrix) const

//-----------SKELETON APPLYONE FUNCTION  - BLOCK DIVISION
//! BEGINApplyUnaryBlock
/*!
  Used to code the user function for the skeleton ApplyUnary in the case of a block distribution
*/
#define BEGINApplyUnaryBlock(name, input_matrix, input_type, r, output_matrix, output_type, r2) struct _ApplyUnary_##name : public _ApplyUnary_Func<input_type,r, output_type,r2,false> \
{ \
  void operator()(DMatrix<input_type,r,false>& input_matrix, DMatrix<output_type,r2,false>& output_matrix) const

//-----------SKELETON APPLYTWO FUNCTION
//! BEGINApplyBinary
/*!
  Used to code the user function for the skeleton ApplyBinary
*/
#define BEGINApplyBinary(name, input_matrix1, input_type1, input_matrix2, input_type2,r, output_matrix, output_type,r3) struct _ApplyBinary_##name : public _ApplyBinary_Func<input_type1,r,input_type2,output_type,r3> \
{ \
  void operator()(DMatrix<input_type1,r>& input_matrix1,DMatrix<input_type2,r>& input_matrix2, DMatrix<output_type,r3>& output_matrix) const

//-----------SKELETON APPLYTWO FUNCTION - BLOCK DIVISION
//! BEGINApplyBinaryBlock
/*!
  Used to code the user function for the skeleton ApplyBinary in the case of a block distribution
*/
#define BEGINApplyBinaryBlock(name, input_matrix1, input_type1, input_matrix2, input_type2,r, output_matrix, output_type,r3) struct _ApplyBinary_##name : public _ApplyBinary_Func<input_type1,r,input_type2,output_type,r3,false> \
{ \
  void operator()(DMatrix<input_type1,r,false>& input_matrix1,DMatrix<input_type2,r,false>& input_matrix2, DMatrix<output_type,r3,false>& output_matrix) const

//-----------SKELETON APPLYLIST FUNCTION
//! BEGINApplyList
/*!
  Used to code the user function for the skeleton ApplyList
*/
#define BEGINApplyList(name, inputs, inputs_type,r, outputs, outputs_type,r2) struct _ApplyList_##name : public _ApplyList_Func<inputs_type,r,outputs_type,r2> \
{ \
  void operator()(DMatrix_vector<inputs_type,r>& inputs, DMatrix_vector<outputs_type,r2>& outputs) const

//-----------SKELETON APPLYLIST FUNCTION - BLOCK DIVISION
//! BEGINApplyListBlock
/*!
  Used to code the user function for the skeleton ApplyList in the case of a block distribution
*/
#define BEGINApplyListBlock(name, inputs, inputs_type,r, outputs, outputs_type,r2) struct _ApplyList_##name : public _ApplyList_Func<inputs_type,r, outputs_type,r2,false> \
{ \
  void operator()(DMatrix_vector<inputs_type,r,false>& inputs, DMatrix_vector<outputs_type,r2,false>& outputs) const

//-----------SKELETON APPLYREDUCTION FUNCTION
//! BEGINApplyReduction
/*!
  Used to code the user function for the skeleton ApplyReduction
*/
#define BEGINApplyReduction(name, input_matrix, type, r,output) struct _ApplyReduction_##name : public _ApplyReduction_Func<type, r> \
{ \
  void operator()(DMatrix<type,r>& input_matrix, type& output) const

//! BEGINApplyReductionBlock
/*!
  Used to code the user function for the skeleton ApplyReduction in the case of a block distribution
*/
#define BEGINApplyReductionBlock(name, input_matrix, type, r,output) struct _ApplyReduction_##name : public _ApplyReduction_Func<type, r,false> \
{ \
  void operator()(DMatrix<type,r,false>& input_matrix, type& output) const

//-----------SKELETON MAP FUNCTION
//! BEGINMap
/*!
  Used to code the user function for the skeleton Map
*/
#define BEGINMap(name, element, element_type, return_type) struct _Map_##name : public _Map_Func<element_type, return_type> \
{ \
  return_type operator()(element_type element) const
//-----------

//=======================================
//============DDAG SKELETONS=============
//=======================================
//-----------SKELETON APPLY UNARY FUNCTION
//! BEGINApplyUnaryDDAG
/*!
  Used to code the user function for the skeleton ApplyUnaryDDAG
*/
#define BEGINApplyUnaryDDAG(name, DPMap_n1, n1, DPMap_e1, e1, DPMap_n2, n2, DPMap_e2, e2) struct _ApplyUnaryDDAG_##name : public _ApplyUnaryDDAG_Func<DPMap_n1, DPMap_e1, DPMap_n2, DPMap_e2> \
{ \
  void operator()(DPMap_n1& n1, DPMap_e1& e1, DPMap_n2& n2, DPMap_e2& e2) const

//-----------SKELETON APPLY UNARY FUNCTION SPECIALIZATION
//! BEGINApplyUnaryDDAG
/*!
  Used to code the user function for the skeleton ApplyUnaryDDAG
*/
#define BEGINSApplyUnaryDDAG(name, DPMap_n, n1, n2, DPMap_e, e1, e2) struct _ApplyUnaryDDAG_##name : public _ApplyUnaryDDAG_Func<DPMap_n, DPMap_e, DPMap_n, DPMap_e> \
{ \
  void operator()(DPMap_n& n1, DPMap_e& e1, DPMap_n& n2, DPMap_e& e2) const

//-----------END oF SKELETON FUNCTIONS
//! END
/*!
  Used to end any BEGIN macro
*/
#define END(name)  }name;
//-----------


}

#endif
